import multiprocessing
import sys
import warnings
from pathlib import Path

import joblib
import matplotlib
import matplotlib.style as mplstyle
import numpy
import scipy
from addict import Dict
from cadet import H5, Cadet
from sklearn import preprocessing
from sklearn.model_selection import cross_val_score

import CADETMatch.evo as evo
import CADETMatch.util as util
from CADETMatch.cache import cache

mplstyle.use("fast")

matplotlib.use("Agg")

import matplotlib.cm
import matplotlib.pyplot as plt
from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from sklearn.preprocessing import MinMaxScaler

cm_plot = matplotlib.cm.gist_rainbow

import logging

import CADETMatch.kde_generator as kde_generator
import CADETMatch.loggerwriter as loggerwriter


def get_color(idx, max_colors, cmap):
    return cmap(1.0 * float(idx) / max_colors)


saltIsotherms = {
    b"STERIC_MASS_ACTION",
    b"SELF_ASSOCIATION",
    b"MULTISTATE_STERIC_MASS_ACTION",
    b"SIMPLE_MULTISTATE_STERIC_MASS_ACTION",
    b"BI_STERIC_MASS_ACTION",
}

size = 20

plt.rc("font", size=size)  # controls default text sizes
plt.rc("axes", titlesize=size)  # fontsize of the axes title
plt.rc("axes", labelsize=size)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=size)  # fontsize of the tick labels
plt.rc("ytick", labelsize=size)  # fontsize of the tick labels
plt.rc("legend", fontsize=size)  # legend fontsize
plt.rc("figure", titlesize=size)  # fontsize of the figure title
plt.rc("figure", autolayout=True)


def plotTube(cache, chain, kde, scaler):
    results, combinations, probability = processChainForPlots(cache, chain, kde, scaler)

    output_mcmc = cache.settings["resultsDirSpace"] / "mcmc"
    output_mcmc.mkdir(parents=True, exist_ok=True)

    mcmc_h5 = output_mcmc / "mcmc_plots.h5"

    h5 = H5()
    h5.filename = mcmc_h5.as_posix()

    for expName, value in combinations.items():
        exp_name = expName.split("_")[0]
        plot_mcmc(
            output_mcmc,
            value,
            expName,
            "combine",
            cache.target[exp_name]["time"],
            cache.target[exp_name]["value"],
            probability,
        )
        h5.root[expName] = value["data"]
        h5.root["exp_%s_time" % expName] = cache.target[exp_name]["time"]
        h5.root["exp_%s_value" % expName] = cache.target[exp_name]["value"]

    for exp, units in results.items():
        for unitName, unit in units.items():
            for comp, data in unit.items():
                expName = "%s_%s" % (exp, unitName)
                plot_mcmc(
                    output_mcmc,
                    data,
                    expName,
                    comp,
                    cache.target[exp]["time"],
                    cache.target[exp]["value"],
                    probability,
                )
                h5.root["%s_%s" % (expName, comp)] = data["data"]
                h5.root["exp_%s_%s_time" % (expName, comp)] = cache.target[exp_name][
                    "time"
                ]
                h5.root["exp_%s_%s_value" % (expName, comp)] = cache.target[exp_name][
                    "value"
                ]
    
    h5.save(lock=True)


def processChainForPlots(cache, chain, kde, scaler):
    (
        mcmc_selected,
        mcmc_selected_transformed,
        mcmc_selected_score,
        results,
        times,
        mcmc_score,
    ) = genRandomChoice(cache, chain, kde, scaler)

    writeSelected(
        cache, mcmc_selected, mcmc_selected_transformed, mcmc_selected_score, mcmc_score
    )

    results, combinations = processResultsForPlotting(results, times)

    return results, combinations, numpy.exp(mcmc_score)


def processResultsForPlotting(results, times):
    for expName, units in list(results.items()):
        for unitName, unit in list(units.items()):
            for compIdx, compValue in list(unit.items()):
                data = numpy.array(compValue)
                results[expName][unitName][compIdx] = {}
                results[expName][unitName][compIdx]["data"] = data
                results[expName][unitName][compIdx]["mean"] = numpy.mean(data, 0)
                results[expName][unitName][compIdx]["std"] = numpy.std(data, 0)
                results[expName][unitName][compIdx]["min"] = numpy.min(data, 0)
                results[expName][unitName][compIdx]["max"] = numpy.max(data, 0)
                results[expName][unitName][compIdx]["time"] = times[expName]

    combinations = {}
    for expName, units in results.items():
        for unitName, unit in list(units.items()):
            data = numpy.zeros(unit[0]["data"].shape)
            times = unit[0]["time"]
            for compIdx, compValue in list(unit.items()):
                data = data + compValue["data"]
            temp = {}
            temp["data"] = data
            temp["time"] = times
            temp["mean"] = numpy.mean(data, 0)
            temp["std"] = numpy.std(data, 0)
            temp["min"] = numpy.min(data, 0)
            temp["max"] = numpy.max(data, 0)
            comb_name = "%s_%s" % (expName, unitName)
            combinations[comb_name] = temp
    return results, combinations


def fitness(individual):
    return evo.fitness(individual, sys.argv[1])


def genRandomChoice(cache, chain, kde, scaler):
    "want about 1000 items and will be removing about 10% of them"
    size = 4000
    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    scaler_mle = joblib.load(mcmcDir / "kde_prior_scaler.joblib")
    kde_mle = joblib.load(mcmcDir / "kde_prior.joblib")

    individuals_mle = kde_mle.sample(size)
    individuals_scaled = scaler_mle.inverse_transform(individuals_mle)

    map_function = util.getMapFunction()
    fitnesses = map_function(fitness, individuals_scaled)

    results = {}
    times = {}

    mcmc_selected = []
    mcmc_selected_transformed = []
    mcmc_selected_score = []

    for (fit, csv_line, meta_score, result, individual) in fitnesses:
        if result is not None:
            mcmc_selected_score.append(tuple(fit))
            for value in result.values():
                mcmc_selected_transformed.append(tuple(value["cadetValues"]))
                mcmc_selected.append(tuple(value["individual"]))
                break

            for key, value in result.items():
                sims = results.get(key, {})

                sim = value["simulation"]

                outlets = [
                    (name, sim.root.input.model[name].ncomp)
                    for name in cache.target[key]["units_used"]
                ]

                for outlet, ncomp in outlets:
                    units = sims.get(outlet, {})

                    for i in range(ncomp):
                        comps = units.get(i, [])
                        comps.append(
                            sim.root.output.solution[outlet][
                                "solution_outlet_comp_%03d" % i
                            ]
                        )
                        units[i] = comps

                    sims[outlet] = units

                    if key not in times:
                        times[key] = sim.root.output.solution.solution_times

                results[key] = sims
        else:
            multiprocessing.get_logger().info(
                "Failure in random choice: fit: %s  csv_line: %s   result:%s",
                fit,
                csv_line,
                result,
            )

    mcmc_selected = numpy.array(mcmc_selected)
    mcmc_selected_transformed = numpy.array(mcmc_selected_transformed)
    mcmc_selected_score = numpy.array(mcmc_selected_score)

    mcmc_score = kde.score_samples(scaler.transform(mcmc_selected_score))

    return (
        mcmc_selected,
        mcmc_selected_transformed,
        mcmc_selected_score,
        results,
        times,
        mcmc_score,
    )


def writeSelected(
    cache, mcmc_selected, mcmc_selected_transformed, mcmc_selected_score, mcmc_score
):
    mcmc_h5 = Path(cache.settings["resultsDirMCMC"]) / "mcmc_selected.h5"
    h5 = H5()
    h5.filename = mcmc_h5.as_posix()
    h5.root.mcmc_selected = numpy.array(mcmc_selected)
    h5.root.mcmc_selected_transformed = numpy.array(mcmc_selected_transformed)
    h5.root.mcmc_selected_score = numpy.array(mcmc_selected_score)
    h5.root.mcmc_selected_kdescore = numpy.array(mcmc_score)

    h5.save(lock=True)


def plot_mcmc(output_mcmc, value, expName, name, expTime, expValue, probability):
    data = value["data"]
    times = value["time"]
    mean = value["mean"]
    std = value["std"]
    minValues = value["min"]
    maxValues = value["max"]

    plt.figure(figsize=[20, 10])
    plt.plot(times, mean, label="mean")
    plt.fill_between(
        times, mean - std, mean + std, color="green", alpha=0.2, label="high prob"
    )
    plt.fill_between(
        times, minValues, maxValues, color="red", alpha=0.2, label="low prob"
    )
    plt.plot(expTime, expValue, "k", label="exp")
    plt.xlabel("time(s)")
    plt.ylabel("conc(mM)")
    plt.legend()
    plt.savefig(
        str(output_mcmc / ("%s_%s.png" % (expName, name))), bbox_inches="tight", dpi=600
    )
    plt.close()

    scaler = MinMaxScaler()
    probability = numpy.squeeze(scaler.fit_transform(probability.reshape(-1, 1)))

    sort_index = numpy.argsort(probability)

    colors = plt.cm.rainbow(probability)

    plt.figure(figsize=[20, 10])
    for idx in sort_index:
        plt.plot(times, data[idx, :], color=colors[idx])
    plt.plot(expTime, expValue, "k", label="exp")
    plt.xlabel("time(s)")
    plt.ylabel("conc(mM)")
    plt.legend()
    sm = plt.cm.ScalarMappable(cmap=plt.cm.rainbow, norm=plt.Normalize(vmin=0, vmax=1))
    plt.colorbar(sm)
    plt.savefig(
        str(output_mcmc / ("%s_%s_lines.png" % (expName, name))),
        bbox_inches="tight",
        dpi=600,
    )
    plt.close()


def main():
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings["resultsDirLog"], "mcmc_plot_tube.log")
    cache.setup(sys.argv[1])

    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    mcmc_h5 = mcmcDir / "mcmc.h5"

    resultDir = Path(cache.settings["resultsDirBase"])
    result_lock = resultDir / "result.lock"

    mcmc_store = H5()
    mcmc_store.filename = mcmc_h5.as_posix()
    mcmc_store.load(
        paths=[
            "/flat_chain",
        ],
        lock=True
    )

    kde, kde_scaler = kde_generator.getKDE(cache)
    plotTube(cache, mcmc_store.root.flat_chain, kde, kde_scaler)


if __name__ == "__main__":
    main()



#This script generates all the h5 example files
from pathlib import Path
from cadet import H5, Cadet
from addict import Dict
import numpy
import pandas
import json
import shutil
import CADETMatch.util

def create_experiments(defaults):
    config = Dict()
    config.CADETPath = Cadet.cadet_path
    config.resultsDir = 'results'
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = True
    
    parameter1 = Dict()
    parameter1.location = '/input/model/unit_001/COL_DISPERSION'
    parameter1.min = 1e-10
    parameter1.max = 1e-6
    parameter1.component = -1
    parameter1.bound = -1
    parameter1.transform = 'auto'

    parameter2 = Dict()
    parameter2.location = '/input/model/unit_001/COL_POROSITY'
    parameter2.min = 0.2
    parameter2.max = 0.7
    parameter2.component = -1
    parameter2.bound = -1
    parameter2.transform = 'auto'

    config.parameters = [parameter1, parameter2]

    experiment1 = Dict()
    experiment1.name = 'main'
    experiment1.csv = 'dextran.csv'
    experiment1.HDF5 = 'dextran.h5'
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    config.experiments = [experiment1,]

    feature1 = Dict()
    feature1.name = "main_feature"
    feature1.type = 'DextranShape'

    experiment1.scores = [feature1,]

    experiments_dir = defaults.base_dir / "experiments"

    single = experiments_dir / "single"
    single.mkdir(parents=True, exist_ok=True)

    multiple = experiments_dir / "multiple"
    multiple.mkdir(parents=True, exist_ok=True)

    advanced = experiments_dir / "multiple_advanced"
    advanced.mkdir(parents=True, exist_ok=True)

    create_common(single, config)   
    

    #multiple
    experiment1 = Dict()
    experiment1.name = 'main1'
    experiment1.csv = 'dextran1.csv'
    experiment1.HDF5 = 'dextran1.h5'
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    experiment2 = Dict()
    experiment2.name = 'main2'
    experiment2.csv = 'dextran2.csv'
    experiment2.HDF5 = 'dextran2.h5'
    experiment2.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    feature1 = Dict()
    feature1.name = "main_feature"
    feature1.type = 'DextranShape'

    experiment1.scores = [feature1,]
    experiment2.scores = [feature1,]

    config.experiments = [experiment1, experiment2]
     
    create_common(multiple, config)   

    #multiple advanced
    #multiple
    experiment1 = Dict()
    experiment1.name = 'dextran'
    experiment1.csv = 'dextran.csv'
    experiment1.HDF5 = 'dextran.h5'
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    experiment2 = Dict()
    experiment2.name = 'non'
    experiment2.csv = 'non.csv'
    experiment2.HDF5 = 'non.h5'
    experiment2.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    feature1 = Dict()
    feature1.name = "main_feature"
    feature1.type = 'DextranShape'

    feature2 = Dict()
    feature2.name = "main_feature"
    feature2.type = 'Shape'
    feature2.decay = 1

    experiment1.scores = [feature1,]
    experiment2.scores = [feature2,]

    config.experiments = [experiment1, experiment2]

    parameter3 = Dict()
    parameter3.location = '/input/model/unit_001/FILM_DIFFUSION'
    parameter3.min = 1e-9
    parameter3.max = 1e-4
    parameter3.component = 0
    parameter3.bound = 0
    parameter3.transform = 'auto'
    parameter3.experiments = ["non"]

    parameter4 = Dict()
    parameter4.location = '/input/model/unit_001/PAR_POROSITY'
    parameter4.min = 0.2
    parameter4.max = 0.7
    parameter4.component = -1
    parameter4.bound = -1
    parameter4.transform = 'auto'
    parameter4.experiments = ["non"]

    config.parameters = [parameter1, parameter2, parameter3, parameter4]
     
    create_common(advanced, config)

def create_scores(defaults):
    create_shared_scores(defaults)
    create_ceiling(defaults)
    create_fractionation(defaults)
    create_slicing(defaults)

def create_slicing(defaults):
    "create all the scores that have the same config except for the score name"
    config = Dict()
    config.CADETPath = Cadet.cadet_path
    config.resultsDir = 'results'
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = True
    
    parameter1 = Dict()
    parameter1.location = '/input/model/unit_001/COL_DISPERSION'
    parameter1.min = 1e-10
    parameter1.max = 1e-6
    parameter1.component = -1
    parameter1.bound = -1
    parameter1.transform = 'auto'

    parameter2 = Dict()
    parameter2.location = '/input/model/unit_001/COL_POROSITY'
    parameter2.min = 0.2
    parameter2.max = 0.7
    parameter2.component = -1
    parameter2.bound = -1
    parameter2.transform = 'auto'

    config.parameters = [parameter1, parameter2]

    experiment1 = Dict()
    experiment1.name = 'main'
    experiment1.csv = 'dextran.csv'
    experiment1.HDF5 = 'dextran.h5'
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    config.experiments = [experiment1,]

    feature1 = Dict()
    feature1.name = "feature_front"
    feature1.type = 'ShapeFront'
    feature1.start = 250
    feature1.stop = 350

    feature2 = Dict()
    feature2.name = "feature_back"
    feature2.type = 'ShapeBack'
    feature2.start = 330
    feature2.stop = 430

    experiment1.scores = [feature1,feature2]

    scores_dir = defaults.base_dir / "scores" 

    dir = scores_dir / "misc" / "multiple_scores_slicing"

    create_common(dir, config)

def create_fractionation(defaults):
    "create the ceiling"
    config = Dict()
    config.CADETPath = Cadet.cadet_path
    config.resultsDir = 'results'
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = True
    
    parameter1 = Dict()
    parameter1.location = ['/input/model/unit_001/adsorption/LIN_KA','/input/model/unit_001/adsorption/LIN_KD']
    parameter1.minKA = 1e-6
    parameter1.maxKA = 1e-2
    parameter1.minKEQ = 1e-3
    parameter1.maxKEQ = 1e3
    parameter1.component = 0
    parameter1.bound = 0
    parameter1.transform = 'auto_keq'

    parameter2 = Dict()
    parameter2.location = ['/input/model/unit_001/adsorption/LIN_KA','/input/model/unit_001/adsorption/LIN_KD']
    parameter2.minKA = 1e-6
    parameter2.maxKA = 1e-2
    parameter2.minKEQ = 1e-3
    parameter2.maxKEQ = 1e3
    parameter2.component = 1
    parameter2.bound = 0
    parameter2.transform = 'auto_keq'

    config.parameters = [parameter1, parameter2]

    experiment1 = Dict()
    experiment1.name = 'main'
    experiment1.csv = 'data_sum.csv'
    experiment1.HDF5 = 'fraction.h5'
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    config.experiments = [experiment1,]

    feature1 = Dict()
    feature1.name = "sum_signal"
    feature1.type = 'Shape'
    feature1.output_path = ['/output/solution/unit_002/SOLUTION_OUTLET_COMP_000', '/output/solution/unit_002/SOLUTION_OUTLET_COMP_001']
    feature1.csv = 'data_sum.csv'

    feature2 = Dict()
    feature2.name = "fractionation"
    feature2.type = 'fractionationSlide'
    feature2.unit_name = "unit_002"
    feature2.fraction_csv = 'frac.csv'

    experiment1.scores = [feature1,feature2]

    scores_dir = defaults.base_dir / "scores"
    dir = scores_dir / "fractionationSlide"

    create_common(dir, config)


    #modify for fractionationSSE

    dir = scores_dir / "other" / "fractionationSSE"

    config.experiments[0].scores[1].type = "fractionationSSE" 
    config.experiments[0].scores[0].type = "SSE" 

    create_common(dir, config)

    dir = scores_dir / "misc" / "multiple_components"

    config.experiments[0].scores[1].type = "Shape" 
    config.experiments[0].scores[1].output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_001'
    config.experiments[0].scores[1].csv = 'comp1.csv'

    config.experiments[0].scores[0].type = "Shape" 
    config.experiments[0].scores[0].output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'
    config.experiments[0].scores[0].csv = 'comp0.csv'

    create_common(dir, config)

def create_ceiling(defaults):
    "create the ceiling"
    config = Dict()
    config.CADETPath = Cadet.cadet_path
    config.resultsDir = 'results'
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = True
    
    parameter1 = Dict()
    parameter1.location = '/input/model/unit_000/sec_000/CONST_COEFF'
    parameter1.min = 0.1
    parameter1.max = 5.0
    parameter1.component = 0
    parameter1.bound = 0
    parameter1.transform = 'auto'

    config.parameters = [parameter1,]

    experiment1 = Dict()
    experiment1.name = 'main'
    experiment1.csv = 'flat.csv'
    experiment1.HDF5 = 'flat.h5'
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    config.experiments = [experiment1,]

    feature1 = Dict()
    feature1.name = "main_feature"
    feature1.type = 'Ceiling'
    feature1.max_value = 1.0

    experiment1.scores = [feature1,]

    scores_dir = defaults.base_dir / "scores"
    dir = scores_dir / "Ceiling"
    score_name = dir.name

    create_common(dir, config)

def create_shared_scores(defaults):
    "create all the scores that have the same config except for the score name"
    config = Dict()
    config.CADETPath = Cadet.cadet_path
    config.resultsDir = 'results'
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = True
    
    parameter1 = Dict()
    parameter1.location = '/input/model/unit_001/COL_DISPERSION'
    parameter1.min = 1e-10
    parameter1.max = 1e-6
    parameter1.component = -1
    parameter1.bound = -1
    parameter1.transform = 'auto'

    parameter2 = Dict()
    parameter2.location = '/input/model/unit_001/COL_POROSITY'
    parameter2.min = 0.2
    parameter2.max = 0.7
    parameter2.component = -1
    parameter2.bound = -1
    parameter2.transform = 'auto'

    config.parameters = [parameter1, parameter2]

    experiment1 = Dict()
    experiment1.name = 'main'
    experiment1.csv = 'dextran.csv'
    experiment1.HDF5 = 'dextran.h5'
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    config.experiments = [experiment1,]

    feature1 = Dict()
    feature1.name = "main_feature"
    feature1.type = 'DextranShape'

    experiment1.scores = [feature1,]

    dextran_paths = ['DextranShape', 'Shape', 'ShapeBack', 'ShapeFront', 'SSE', 
                     'other/curve', 'other/DextranSSE', 'other/ShapeDecay',
                     'other/ShapeDecayNoDer', 'other/ShapeDecaySimple', 'other/ShapeNoDer', 
                     'other/ShapeOnly', 'other/ShapeSimple', 'other/similarity', 
                     'other/similarityDecay']

    scores_dir = defaults.base_dir / "scores"

    for path in dextran_paths:
        dir = scores_dir / path
        score_name = dir.name
        temp_config = config.deepcopy()
        temp_config.experiments[0].scores[0].type = score_name

        if score_name in ('Shape', 'ShapeBack', 'ShapeFront'):
            temp_config.experiments[0].scores[0].decay = 0
            temp_config.experiments[0].scores[0].derivative = 1

        if score_name in ('ShapeBack', 'ShapeFront'):
            temp_config.experiments[0].scores[0].max_percent = 0.98
            temp_config.experiments[0].scores[0].min_percent = 0.02

        create_common(dir, temp_config)


def create_search(defaults):
    config = Dict()
    config.CADETPath = Cadet.cadet_path
    config.resultsDir = 'results'
    
    parameter1 = Dict()
    parameter1.location = '/input/model/unit_001/COL_DISPERSION'
    parameter1.min = 1e-10
    parameter1.max = 1e-6
    parameter1.component = -1
    parameter1.bound = -1
    parameter1.transform = 'auto'

    parameter2 = Dict()
    parameter2.location = '/input/model/unit_001/COL_POROSITY'
    parameter2.min = 0.2
    parameter2.max = 0.7
    parameter2.component = -1
    parameter2.bound = -1
    parameter2.transform = 'auto'

    config.parameters = [parameter1, parameter2]

    experiment1 = Dict()
    experiment1.name = 'main'
    experiment1.csv = 'dextran.csv'
    experiment1.HDF5 = 'dextran.h5'
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    config.experiments = [experiment1,]

    feature1 = Dict()
    feature1.name = "main_feature"
    feature1.type = 'DextranShape'

    experiment1.scores = [feature1,]

    create_nsga3(defaults, config)
    create_unsga3(defaults, config)
    create_multistart(defaults, config)
    create_graphspace(defaults, config)
    create_scoretest(defaults, config)
    create_gradient(defaults, config)
    create_early_stopping(defaults, config)
    create_refine_shape(defaults, config)
    create_refine_sse(defaults, config)
    create_altScore(defaults, config)
    create_mcmc_stage1(defaults, config)
    create_mcmc_stage2(defaults, config)

def create_common(dir, config, altDir=None):
    if altDir is not None:
        config.baseDir = altDir.as_posix()
    else:
        config.baseDir = dir.as_posix()

    match_config_file = dir / 'dextran.json'

    with open(match_config_file.as_posix(), 'w') as json_file:
        json.dump(config.to_dict(), json_file, indent='\t')

def create_mcmc_stage1(defaults, config):
    dir = defaults.base_dir / "search" / "mcmc" / "stage1"
    config = config.deepcopy()
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.continueMCMC = 1
    config.MCMCpopulationSet = defaults.MCMCpopulation

    error_model = Dict()
    error_model.file_path = "dextran.h5"
    error_model.experimental_csv = "dextran.csv"
    error_model.name = "main"
    error_model.units = [2]
    error_model.delay = [0.0, 2.0]
    error_model.flow = [1.0, 0.001]
    error_model.load = [1.0, 0.001]
    error_model.uv_noise_norm = [1.0, 0.001]

    config.errorModelCount = 1000
    config.errorModel = [error_model,]
    create_common(dir, config)

def create_mcmc_stage2(defaults, config):
    dir = defaults.base_dir / "search" / "mcmc" / "stage2"
    config = config.deepcopy()
    config.baseDir = dir.as_posix()
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.continueMCMC = 1
    config.MCMCpopulationSet = defaults.MCMCpopulation

    error_model = Dict()
    error_model.file_path = "non.h5"
    error_model.experimental_csv = "non.csv"
    error_model.name = "main"
    error_model.units = [2]
    error_model.delay = [0.0, 2.0]
    error_model.flow = [1.0, 0.001]
    error_model.load = [1.0, 0.001]
    error_model.uv_noise_norm = [1.0, 0.001]

    config.errorModelCount = 1000
    config.errorModel = [error_model,]

    config.experiments[0].csv = "non.csv"
    config.experiments[0].HDF5 = "non.h5"

    config.experiments[0].scores[0].type = "Shape"
    config.experiments[0].scores[0].decay = 1

    parameter1 = Dict()
    parameter1.location = '/input/model/unit_001/FILM_DIFFUSION'
    parameter1.min = 1e-12
    parameter1.max = 1e-2
    parameter1.component = -1
    parameter1.bound = -1
    parameter1.transform = 'auto'

    parameter2 = Dict()
    parameter2.location = '/input/model/unit_001/PAR_POROSITY'
    parameter2.min = 0.2
    parameter2.max = 0.7
    parameter2.component = -1
    parameter2.bound = -1
    parameter2.transform = 'auto'

    config.parameters = [parameter1, parameter2]

    mcmc_h5 = Path(defaults.base_dir / "search" / "mcmc" / "stage1" / "results" / "mcmc_refine" / "mcmc" / "mcmc.h5")
    config.mcmc_h5 = mcmc_h5.as_posix()

    match_config_file = dir / 'non.json'

    with open(match_config_file.as_posix(), 'w') as json_file:
        json.dump(config.to_dict(), json_file, indent='\t')

def create_altScore(defaults, config):
    dir = defaults.base_dir / "search" / "other" / "altScore"
    dir.mkdir(parents=True, exist_ok=True)
    config = config.deepcopy()
    config.searchMethod = 'AltScore'
    config.population = 0
    config.PreviousResults = (defaults.base_dir / "search" / "nsga3" / "results" / "result.h5").as_posix()
    config.experiments[0].scores[0].type = "Shape"
    config.resultsDir = (dir / "results").as_posix()
    config.resultsDirOriginal = (defaults.base_dir / "search" / "nsga3" / "results").as_posix()
    create_common(dir, config, altDir = (defaults.base_dir / "search" / "nsga3"))

def create_refine_shape(defaults, config):
    dir = defaults.base_dir / "search" / "other" / "refine_shape"
    config = config.deepcopy()
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = False
    create_common(dir, config)

def create_refine_sse(defaults, config):
    dir = defaults.base_dir / "search" / "other" / "refine_sse"
    config = config.deepcopy()
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = True
    create_common(dir, config)

def create_early_stopping(defaults, config):
    dir = defaults.base_dir / "search" / "other" / "early_stopping"
    config = config.deepcopy()
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = True
    config.stopAverage = 1e-3
    config.stopBest = 1e-3
    create_common(dir, config)

def create_gradient(defaults, config):
    dir = defaults.base_dir / "search" / "gradient"
    config = config.deepcopy()
    config.searchMethod = 'Gradient'
    config.population = 0
    config.gradVector = 1
    config.experiments[0].scores[0].type = "SSE"
    config.seeds = [[2e-7, 0.37],[1e-7, 0.37],[2e-7, 0.41],[1e-6, 0.5],[1e-10, 0.2]]

    create_common(dir, config)

def create_scoretest(defaults, config):
    dir = defaults.base_dir / "search" / "scoretest"
    config = config.deepcopy()
    config.searchMethod = 'ScoreTest'
    config.population = 0
    config.seeds = [[2e-7, 0.37],[1e-7, 0.37],[2e-7, 0.41],[1e-6, 0.5]]

    create_common(dir, config)

def create_graphspace(defaults, config):
    dir = defaults.base_dir / "search" / "graphSpace"
    config = config.deepcopy()
    config.searchMethod = 'GraphSpace'
    config.population = defaults.population
    config.gradVector = True

    create_common(dir, config)

def create_multistart(defaults, config):
    dir = defaults.base_dir / "search" / "multistart"
    config = config.deepcopy()
    config.searchMethod = 'Multistart'
    config.population = defaults.population
    config.gradVector = True
    create_common(dir, config)

def create_nsga3(defaults, config):
    dir = defaults.base_dir / "search" / "nsga3"
    config = config.deepcopy()
    config.searchMethod = 'NSGA3'
    config.population = defaults.population
    config.stallGenerations = 10
    config.finalGradRefinement = True
    config.gradVector = True
    create_common(dir, config)

def create_unsga3(defaults, config):
    dir = defaults.base_dir / "search" / "unsga3"
    config = config.deepcopy()
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.stallGenerations = 10
    config.finalGradRefinement = True
    config.gradVector = True
    create_common(dir, config)

def create_transforms(defaults):
    create_transforms_dextran(defaults)
    create_transforms_non(defaults)
    create_experiments_linear(defaults)
    create_experiments_cstr(defaults)
    create_experiments_linear_exp(defaults)
    create_experiments_index(defaults)

def create_experiments_index(defaults):
    config = Dict()
    config.CADETPath = Cadet.cadet_path
    config.resultsDir = 'results'
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = True
    
    parameter1 = Dict()
    parameter1.location = '/input/model/unit_001/adsorption/LIN_KA'
    parameter1.min = 1e-6
    parameter1.max = 1e-2
    parameter1.index = 0
    parameter1.transform = 'auto'

    parameter2 = Dict()
    parameter2.location = '/input/model/unit_001/adsorption/LIN_KA'
    parameter2.min = 1e-6
    parameter2.max = 1e-2
    parameter2.index = 1
    parameter2.transform = 'auto'

    config.parameters = [parameter1, parameter2]

    experiment1 = Dict()
    experiment1.name = 'main'
    experiment1.csv = 'data_sum.csv'
    experiment1.HDF5 = 'fraction.h5'
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    config.experiments = [experiment1,]

    feature1 = Dict()
    feature1.name = "comp0"
    feature1.type = 'Shape'
    feature1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'
    feature1.csv = 'comp0.csv'

    feature2 = Dict()
    feature2.name = "comp1"
    feature2.type = 'Shape'
    feature2.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_001'
    feature2.csv = 'comp1.csv'

    experiment1.scores = [feature1,feature2]

    create_common(defaults.base_dir / "transforms" / "misc" / "index", config)

def create_experiments_linear_exp(defaults):
    config = Dict()
    config.CADETPath = Cadet.cadet_path
    config.resultsDir = 'results'
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = True
    
    parameter1 = Dict()
    parameter1.location = '/input/model/unit_001/COL_DISPERSION'
    parameter1.minLower = 1e-8
    parameter1.maxLower = 1e-6
    parameter1.minUpper = 1e-8
    parameter1.maxUpper = 1e-6
    parameter1.minX = 1
    parameter1.maxX = 3
    parameter1.x_name = 'pH'
    parameter1.component = -1
    parameter1.bound = -1
    parameter1.transform = 'norm_linear'
    
    config.parameters = [parameter1]

    experiment1 = Dict()
    experiment1.name = 'main1'
    experiment1.csv = 'dex1.csv'
    experiment1.HDF5 = 'dex1.h5'
    experiment1.pH = 1
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    experiment2 = experiment1.deepcopy()
    experiment2.name = 'main2'
    experiment2.csv = 'dex2.csv'
    experiment2.HDF5 = 'dex2.h5'
    experiment2.pH = 2

    experiment3 = experiment1.deepcopy()
    experiment3.name = 'main3'
    experiment3.csv = 'dex3.csv'
    experiment3.HDF5 = 'dex3.h5'
    experiment3.pH = 3

    config.experiments = [experiment1, experiment2, experiment3]

    feature1 = Dict()
    feature1.name = "main_feature"
    feature1.type = 'Shape'
    feature1.decay = 1

    experiment1.scores = [feature1,]
    experiment2.scores = [feature1,]
    experiment3.scores = [feature1,]

    experiments_dir = defaults.base_dir / "transforms"

    create_common(experiments_dir / "norm_linear", config)

    parameter1.transform = 'linear'

    create_common(experiments_dir / "other" / "linear", config)

def create_experiments_cstr(defaults):
    config = Dict()
    config.CADETPath = Cadet.cadet_path
    config.resultsDir = 'results'
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = True
    
    parameter1 = Dict()
    parameter1.location = '/input/model/unit_001/INIT_VOLUME'
    parameter1.min = 1e-7
    parameter1.max = 1e-4
    parameter1.component = -1
    parameter1.bound = -1
    parameter1.transform = 'auto'

    parameter2 = Dict()
    parameter2.locationFrom = '/input/model/unit_001/INIT_VOLUME'
    parameter2.componentFrom = -1
    parameter2.boundFrom = -1
    parameter2.locationTo = '/input/model/unit_003/INIT_VOLUME'
    parameter2.componentTo = -1
    parameter2.boundTo = -1
    parameter2.transform = 'set_value'

    parameter3 = Dict()
    parameter3.location1 = '/input/model/unit_001/INIT_VOLUME'
    parameter3.component1 = -1
    parameter3.bound1 = -1
    parameter3.location2 = '/input/model/unit_003/INIT_VOLUME'
    parameter3.component2 = -1
    parameter3.bound2 = -1
    parameter3.locationSum = '/input/model/unit_004/INIT_VOLUME'
    parameter3.componentSum = -1
    parameter3.boundSum = -1
    parameter3.transform = 'sum'

    config.parameters = [parameter1,parameter2, parameter3]

    experiment1 = Dict()
    experiment1.name = 'main'
    experiment1.csv = 'cstr.csv'
    experiment1.HDF5 = 'cstr.h5'
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    config.experiments = [experiment1,]

    feature1 = Dict()
    feature1.name = "main_feature"
    feature1.type = 'Shape'
    feature1.decay = 1

    experiment1.scores = [feature1,]

    experiments_dir = defaults.base_dir / "transforms"

    create_common(experiments_dir / "sum", config)

def create_experiments_linear(defaults):
    config = Dict()
    config.CADETPath = Cadet.cadet_path
    config.resultsDir = 'results'
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = True
    
    parameter1 = Dict()
    parameter1.location = ['/input/model/unit_001/adsorption/LIN_KA','/input/model/unit_001/adsorption/LIN_KD']
    parameter1.minKA = 1e-5
    parameter1.maxKA = 1e-3
    parameter1.minKEQ = 1e-2
    parameter1.maxKEQ = 1e2
    parameter1.component = 0
    parameter1.bound = 0
    parameter1.transform = 'auto_keq'

    config.parameters = [parameter1,]

    experiment1 = Dict()
    experiment1.name = 'main'
    experiment1.csv = 'lin.csv'
    experiment1.HDF5 = 'lin.h5'
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    config.experiments = [experiment1,]

    feature1 = Dict()
    feature1.name = "main_feature"
    feature1.type = 'Shape'
    feature1.decay = 1

    experiment1.scores = [feature1,]

    experiments_dir = defaults.base_dir / "transforms"

    linear_paths = ['auto_keq', 'other/keq', 'other/norm_keq']

    for path in linear_paths:
        dir = experiments_dir / path
        score_name = dir.name
        temp_config = config.deepcopy()
        temp_config.parameters[0].transform = score_name

        create_common(dir, temp_config)

    #set_value estimate one concentration and set the other
    temp_config = config.deepcopy()

    parameter1 = Dict()
    parameter1.location = '/input/model/unit_000/sec_000/CONST_COEFF'
    parameter1.min = 1e-2
    parameter1.max = 1e0
    parameter1.component = 0
    parameter1.bound = 0
    parameter1.transform = 'auto'

    parameter2 = Dict()
    parameter2.locationFrom = '/input/model/unit_000/sec_000/CONST_COEFF'
    parameter2.componentFrom = 0
    parameter2.boundFrom = 0
    parameter2.locationTo = '/input/model/unit_000/sec_000/CONST_COEFF'
    parameter2.componentTo = 1
    parameter2.boundTo = 0
    parameter2.transform = 'set_value'

    temp_config.parameters = [parameter1,parameter2]
    create_common(experiments_dir / "set_value", temp_config)


def create_transforms_non(defaults):
    config = Dict()
    config.CADETPath = Cadet.cadet_path
    config.resultsDir = 'results'
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = True
    
    parameter1 = Dict()
    parameter1.location = '/input/model/unit_001/FILM_DIFFUSION'
    parameter1.min = 1e-8
    parameter1.max = 1e-4
    parameter1.component = 0
    parameter1.bound = 0
    parameter1.transform = 'auto_inverse'

    config.parameters = [parameter1,]

    experiment1 = Dict()
    experiment1.name = 'main'
    experiment1.csv = 'non.csv'
    experiment1.HDF5 = 'non.h5'
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    config.experiments = [experiment1,]

    feature1 = Dict()
    feature1.name = "main_feature"
    feature1.type = 'Shape'
    feature1.decay = 1

    experiment1.scores = [feature1,]

    experiments_dir = defaults.base_dir / "transforms"
    create_common(experiments_dir / "auto_inverse", config)

    #
    parameter1 = Dict()
    parameter1.locationFrom = '/input/model/unit_001/COL_POROSITY'
    parameter1.componentFrom= -1
    parameter1.boundFrom = -1
    parameter1.locationTo = '/input/model/unit_001/PAR_POROSITY'
    parameter1.componentTo = -1
    parameter1.boundTo = -1
    parameter1.min = -0.1
    parameter1.max = 0.1    
    parameter1.transform = 'norm_add'

    config.parameters = [parameter1,]
    create_common(experiments_dir / "norm_add", config)


    parameter1 = Dict()
    parameter1.locationFrom = '/input/model/unit_001/COL_POROSITY'
    parameter1.componentFrom= -1
    parameter1.boundFrom = -1
    parameter1.locationTo = '/input/model/unit_001/PAR_POROSITY'
    parameter1.componentTo = -1
    parameter1.boundTo = -1
    parameter1.min = 0.8
    parameter1.max = 1.5    
    parameter1.transform = 'norm_mult'

    config.parameters = [parameter1,]
    create_common(experiments_dir / "norm_mult", config)

def create_transforms_dextran(defaults):
    config = Dict()
    config.CADETPath = Cadet.cadet_path
    config.resultsDir = 'results'
    config.searchMethod = 'UNSGA3'
    config.population = defaults.population
    config.gradVector = True
    
    parameter1 = Dict()
    parameter1.location = '/input/model/unit_001/COL_POROSITY'
    parameter1.min = 0.2
    parameter1.max = 0.7
    parameter1.component = -1
    parameter1.bound = -1
    parameter1.transform = 'auto'

    config.parameters = [parameter1,]

    experiment1 = Dict()
    experiment1.name = 'main'
    experiment1.csv = 'dextran.csv'
    experiment1.HDF5 = 'dextran.h5'
    experiment1.output_path = '/output/solution/unit_002/SOLUTION_OUTLET_COMP_000'

    config.experiments = [experiment1,]

    feature1 = Dict()
    feature1.name = "main_feature"
    feature1.type = 'DextranShape'

    experiment1.scores = [feature1,]

    experiments_dir = defaults.base_dir / "transforms"

    dextran_paths = ['auto', 'other/log', 'other/norm', 'other/norm_log', 'other/null']

    for path in dextran_paths:
        dir = experiments_dir / path
        score_name = dir.name
        temp_config = config.deepcopy()
        temp_config.parameters[0].transform = score_name

        create_common(dir, temp_config)

    temp_config = config.deepcopy()

    parameter1 = Dict()
    parameter1.location = '/input/model/unit_001/CROSS_SECTION_AREA'
    parameter1.min = 1e-3
    parameter1.max = 1e-1
    parameter1.component = -1
    parameter1.bound = -1
    parameter1.transform = 'norm_diameter'

    temp_config.parameters = [parameter1,]

    create_common(experiments_dir / 'norm_diameter', temp_config)

    parameter1.transform = 'diameter'

    create_common(experiments_dir / 'other/diameter', temp_config)


    parameter1 = Dict()
    parameter1.area_location = '/input/model/unit_001/CROSS_SECTION_AREA'
    parameter1.length_location = '/input/model/unit_001/COL_LENGTH'
    parameter1.minVolume = 1e-6
    parameter1.maxVolume = 1e-4
    parameter1.minArea = 1e-5
    parameter1.maxArea = 1e-3
    parameter1.component = -1
    parameter1.bound = -1
    parameter1.transform = 'norm_volume_area'

    temp_config.parameters = [parameter1,]

    create_common(experiments_dir / 'norm_volume_area', temp_config)

    parameter1.transform = 'volume_area'

    create_common(experiments_dir / 'other/volume_area', temp_config)


    parameter1 = Dict()
    parameter1.area_location = '/input/model/unit_001/CROSS_SECTION_AREA'
    parameter1.length_location = '/input/model/unit_001/COL_LENGTH'
    parameter1.minVolume = 1e-6
    parameter1.maxVolume = 1e-4
    parameter1.minLength = 0.1
    parameter1.maxLength = 0.3
    parameter1.component = -1
    parameter1.bound = -1
    parameter1.transform = 'norm_volume_length'

    temp_config.parameters = [parameter1,]

    create_common(experiments_dir / 'norm_volume_length', temp_config)

    parameter1.transform = 'volume_length'

    create_common(experiments_dir / 'other/volume_length', temp_config)


def main(defaults):
    "create simulations by directory"
    create_experiments(defaults)
    create_scores(defaults)
    create_search(defaults)
    create_transforms(defaults)

if __name__ == "__main__":
    main()




import create_examples
import create_example_config
import subprocess
import sys
from pathlib import Path
from cadet import H5, Cadet
from addict import Dict

defaults = Dict()
defaults.cadet_path = Path(sys.argv[2]).as_posix()
defaults.base_dir = Path(sys.argv[1]).resolve()
defaults.flow_rate = 2.88e-8 # m^3/s
defaults.ncol = 100
defaults.npar = 10
defaults.abstol = 1e-8
defaults.algtol = 1e-10
defaults.reltol = 1e-8
defaults.lin_ka1 = 4e-4
defaults.lin_ka2 = 1e-4
defaults.lin_kd1 = 4e-3
defaults.lin_kd2 = 1e-3
defaults.col_dispersion = 2e-7
defaults.film_diffusion = 1e-6
defaults.par_diffusion = 3e-11

population = int(sys.argv[3])
mcmc_population = int(sys.argv[4])

if population:
    defaults.population = population
else:
    defaults.population = 20

if mcmc_population:
    defaults.MCMCpopulation = mcmc_population
else:
    defaults.MCMCpopulation = 20

Cadet.cadet_path = defaults.cadet_path

if __name__ == "__main__":
    create_examples.main(defaults)
    create_example_config.main(defaults)




import argparse
import importlib
import subprocess
import sys
import pathlib

import psutil


def makeParser():
    """Create the CADETMatch module arguments parser."""
    parser = argparse.ArgumentParser(
        description="Starts a parallel version of match using multiprocessing.",
        prog="{0} -m CADETMatch".format(sys.executable),
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--json", "-j", help="Path to JSON file", metavar="JSON")
    parser.add_argument(
        "--match",
        help="Run CADETMatch with Parameter Estimation and Error Modeling",
        action="store_true",
    )

    parser.add_argument(
        "--generate_corner", help="Generate corner graphs", action="store_true"
    )

    parser.add_argument(
        "--generate_graphs",
        help="Generate general graphs excluding 3D",
        action="store_true",
    )

    parser.add_argument(
        "--generate_graphs_all",
        help="Generate general graphs including 3D",
        action="store_true",
    )

    parser.add_argument(
        "--generate_graphs_autocorr",
        help="Generate autocorrelation graphs",
        action="store_true",
    )

    parser.add_argument(
        "--generate_mcmc_plot_tube",
        help="Generate mcmc plot tube graphs",
        action="store_true",
    )

    parser.add_argument(
        "--generate_mcmc_mixing",
        help="Generate mcmc mixing graphs",
        action="store_true",
    )

    parser.add_argument(
        "--generate_mle", "--generate_map", help="Generate maximum a posteriori", action="store_true"
    )

    parser.add_argument(
        "-n", help="Number of parallel processes to use. Use 1 for debugging", default=0
    )

    parser.add_argument(
        "--generate_examples", help="Directory for CADETMatch Example generation", action="store", 
        type=pathlib.Path
    )

    parser.add_argument(
        "--cadet_examples", help="Set path to the cadet-cli binary for examples to use", action="store", 
        type=pathlib.Path
    )

    parser.add_argument(
        "--example_population", help="Set population per estimated variable", action="store",
        default=0,
        type=int
    )

    parser.add_argument(
        "--example_mcmc_population", help="Set mcmc population to use", action="store", 
        default=0,
        type=int
    )

    parser.add_argument(
        "--run_examples", help="Directory for CADETMatch Example running", action="store", 
        type=pathlib.Path
    )

    parser.add_argument(
        "--results_examples", help="Gather results from directory for CADETMatch Example that have been run", action="store", 
        type=pathlib.Path
    )

    parser.add_argument(
        "--clean_examples", help="Directory for CADETMatch Example cleaning (remove results directories)", action="store", 
        type=pathlib.Path
    )    

    return parser


def run_command(module, json, number_of_jobs, additional=None):
    command = [
        sys.executable,
    ]
    command.extend([importlib.util.find_spec(module).origin, str(json)])
    if additional is not None:
        command.extend(additional)

    command.append(str(number_of_jobs))

    rc = subprocess.run(command, bufsize=1)

    return rc.returncode

def run_examples(module, *args):
    command = [
        sys.executable,
    ]
    command.extend([importlib.util.find_spec(module).origin, ])

    for arg in args:
        command.append(str(arg))

    rc = subprocess.run(command, bufsize=1)

    return rc.returncode


if __name__ == "__main__":
    parser = makeParser()
    args = parser.parse_args()

    if args.generate_examples and args.cadet_examples is None:
        parser.error("--generate_examples requires --cadet_examples")

    if args.match:
        sys.exit(run_command("CADETMatch.match", args.json, args.n))
    if args.generate_corner:
        sys.exit(run_command("CADETMatch.generate_corner_graphs", args.json, args.n))
    if args.generate_graphs_autocorr:
        sys.exit(run_command("CADETMatch.generate_autocorr_graphs", args.json, args.n))
    if args.generate_mcmc_mixing:
        sys.exit(run_command("CADETMatch.generate_mixing_graphs", args.json, args.n))
    if args.generate_graphs:
        sys.exit(run_command("CADETMatch.generate_graphs", args.json, args.n, ["1"]))
    if args.generate_graphs_all:
        sys.exit(run_command("CADETMatch.generate_graphs", args.json, args.n, ["2"]))
    if args.generate_mcmc_plot_tube:
        sys.exit(run_command("CADETMatch.mcmc_plot_tube", args.json, args.n))
    if args.generate_mle:
        sys.exit(run_command("CADETMatch.mle", args.json, args.n))
    if args.clean_examples:
        sys.exit(run_examples("CADETMatch.clean_examples", args.clean_examples))
    if args.run_examples:
        sys.exit(run_examples("CADETMatch.run_examples", args.run_examples, args.n))
    if args.results_examples:
        sys.exit(run_examples("CADETMatch.results_examples", args.results_examples))
    if args.generate_examples:
        sys.exit(run_examples("CADETMatch.generate_examples", args.generate_examples, args.cadet_examples,
                              args.example_population, args.example_mcmc_population))



import subprocess
import sys
from pathlib import Path

def run_matching():
    for path in sorted(Path(sys.argv[1]).resolve().rglob("*.json")):
        if not (path.parent / "results").exists() and path.parent.name != "results" and path.parent.name != "mcmc_refine":
            print(path)
            command = [sys.executable, '-m', 'CADETMatch', '--match', '-j', path.as_posix(), '-n', sys.argv[2]]
            #print(command)
            subprocess.run(command)

if __name__ == "__main__":
    run_matching()



import itertools
import multiprocessing
import time

import numpy
import SALib.sample.sobol_sequence
from sklearn import preprocessing
from sklearn.base import clone
from sklearn.model_selection import GridSearchCV, cross_val_score
from sklearn.neighbors import KernelDensity

import CADETMatch.util as util


def fit(x):
    bw, kde_params, scores = x
    kde = KernelDensity(**kde_params)
    kde.set_params(bandwidth=bw)
    cross_val = cross_val_score(kde, scores, cv=20)
    return cross_val, bw

def get_bandwidth(kde, scores):
    start = time.time()
    map = util.getMapFunction()

    bw = numpy.logspace(-3, 0, 60)

    results = list(
        map(fit, zip(bw, itertools.repeat(kde.get_params()), itertools.repeat(scores)))
    )

    cross_val, bw = zip(*results)

    cross_val = numpy.array(cross_val)
    bw = numpy.array(bw)

    sort_index = numpy.argsort(bw)

    bw = bw[sort_index]
    cross_val = cross_val[sort_index, :]

    scaler = preprocessing.MinMaxScaler()
    shape = cross_val.shape
    cross_val = scaler.fit_transform(cross_val.reshape(-1, 1))
    cross_val = cross_val.reshape(shape)

    mean_value = numpy.mean(cross_val, 1)

    bandwidth = interp_bw(bw, mean_value, cross_val)

    multiprocessing.get_logger().info(
        "Bandwidth %s found in %s seconds", bandwidth, time.time() - start
    )

    store = numpy.array([bw, mean_value]).T

    kde_final = clone(kde)
    kde_final.set_params(bandwidth=bandwidth)

    return kde_final, bandwidth, store


def interp_bw(params, mean_value, values):
    index = numpy.argmax(mean_value)
    params = numpy.array(params)
    mean_value = numpy.array(mean_value)

    indexes = numpy.array([index - 1, index, index + 1])
    x = params[indexes]

    X = numpy.tile(x.reshape(-1, 1), values.shape[1])
    Y = values[indexes, :]

    X = numpy.squeeze(X.reshape(-1, 1))
    Y = numpy.squeeze(Y.reshape(-1, 1))

    poly, res = numpy.polynomial.Polynomial.fit(X, Y, 2, full=True)

    bw = poly.deriv().roots()[0]

    return bw



import sys

import matplotlib
import matplotlib.style as mplstyle

mplstyle.use("fast")

matplotlib.use("Agg")

size = 20

matplotlib.rc("font", size=size)  # controls default text sizes
matplotlib.rc("axes", titlesize=size)  # fontsize of the axes title
matplotlib.rc("axes", labelsize=size)  # fontsize of the x and y labels
matplotlib.rc("xtick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("ytick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("legend", fontsize=size)  # legend fontsize
matplotlib.rc("figure", titlesize=size)  # fontsize of the figure title
matplotlib.rc("figure", autolayout=True)

import itertools
import logging

# parallelization
import multiprocessing
import os
import warnings
from pathlib import Path

import numpy
import pandas
import scipy.interpolate
from addict import Dict
from cadet import H5, Cadet
from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from mpl_toolkits.mplot3d import Axes3D

import CADETMatch.loggerwriter as loggerwriter
import CADETMatch.smoothing as smoothing
import CADETMatch.util as util
from CADETMatch.cache import cache

saltIsotherms = {
    b"STERIC_MASS_ACTION",
    b"SELF_ASSOCIATION",
    b"MULTISTATE_STERIC_MASS_ACTION",
    b"SIMPLE_MULTISTATE_STERIC_MASS_ACTION",
    b"BI_STERIC_MASS_ACTION",
}


import matplotlib.cm
from matplotlib.colors import ListedColormap

cmap = matplotlib.cm.winter

# Get the colormap colors
my_cmap = cmap(numpy.arange(cmap.N))

# Set alpha
my_cmap[:, -1] = 1.0

# Create new colormap
my_cmap = ListedColormap(my_cmap)

cm_plot = matplotlib.cm.gist_rainbow


def get_color(idx, max_colors, cmap):
    return cmap(1.0 * float(idx) / max_colors)


def main(map_function):
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings["resultsDirLog"], "graph.log")
    cache.setup(sys.argv[1])

    multiprocessing.get_logger().info("graphing directory %s", os.getcwd())

    # full generation 1 = 2D, 2 = 2D + 3D
    fullGeneration = int(sys.argv[2])

    # load all data, all data that can be changed by the main process MUST be loaded and this point and the files locked
    # this allows the main process to wait for data loading and then continue once complete

    resultDir = Path(cache.settings["resultsDirBase"])
    progress = resultDir / "progress.csv"
    result_lock = resultDir / "result.lock"

    result_h5 = resultDir / "result.h5"

    progress_df = pandas.read_csv(progress)

    result_data = H5()
    result_data.filename = result_h5.as_posix()
    result_data.load(
        paths=[
            "/output",
            "/output_meta",
            "/is_extended_input",
            "/probability",
            "/input_transform_extended",
            "/input_transform",
            "/mean",
            "/confidence",
            "/distance_correct",
        ],
        lock=True
    )

    graphMeta(cache, map_function)
    graphProgress(cache, map_function, progress_df)

    if fullGeneration:
        graphSpace(fullGeneration, cache, map_function, result_data)
        graphExperiments(cache, map_function)


def graphDistance(cache, map_function, data):
    output_distance = cache.settings["resultsDirSpace"] / "distance"
    output_distance_meta = output_distance / "meta"

    output_distance.mkdir(parents=True, exist_ok=True)

    output_distance_meta.mkdir(parents=True, exist_ok=True)

    parameter_headers_actual = cache.parameter_headers_actual
    score_headers = cache.score_headers
    meta_headers = cache.meta_headers

    temp = []

    if (
        "mean" in data.root
        and "confidence" in data.root
        and "distance_correct" in data.root
    ):
        for idx_parameter, parameter in enumerate(parameter_headers_actual):
            for idx_score, score in enumerate(score_headers):
                temp.append(
                    (
                        output_distance,
                        parameter,
                        idx_parameter,
                        score,
                        idx_score,
                        data.root.distance_correct[:, idx_parameter],
                        data.root.output[:, idx_score],
                        data.root.mean[:, idx_parameter],
                        data.root.confidence[:, idx_parameter],
                    )
                )

        for idx_parameter, parameter in enumerate(parameter_headers_actual):
            for idx_score, score in enumerate(meta_headers):
                temp.append(
                    (
                        output_distance_meta,
                        parameter,
                        idx_parameter,
                        score,
                        idx_score,
                        data.root.distance_correct[:, idx_parameter],
                        data["output_meta"][:, idx_score],
                        data.root.mean[:, idx_parameter],
                        data.root.confidence[:, idx_parameter],
                    )
                )

        list(map_function(plot_2d_scatter, temp))


def plot_2d_scatter(args):
    (
        output_directory,
        parameter,
        parameter_idx,
        score,
        score_idx,
        parameter_data,
        score_data,
        mean_data,
        confidence_data,
    ) = args
    fig = figure.Figure()
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)
    graph.scatter(parameter_data, score_data, c=score_data, cmap=my_cmap)
    graph.set_xlabel(parameter)
    graph.set_ylabel(score)
    filename = "%s_%s.png" % (parameter_idx, score_idx)
    fig.savefig(str(output_directory / filename))

    fig = figure.Figure()
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)
    graph.scatter(mean_data, score_data, c=score_data, cmap=my_cmap)
    graph.set_xlabel(parameter)
    graph.set_ylabel(score)
    filename = "%s_%s_mean.png" % (parameter_idx, score_idx)
    fig.savefig(str(output_directory / filename))

    fig = figure.Figure()
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)
    graph.scatter(confidence_data, score_data, c=score_data, cmap=my_cmap)
    graph.set_xlabel(parameter)
    graph.set_ylabel(score)
    filename = "%s_%s_confidence.png" % (parameter_idx, score_idx)
    fig.savefig(str(output_directory / filename))


def graphExperiments(cache, map_function):
    multiprocessing.get_logger().info("starting simulation graphs")
    directory = Path(cache.settings["resultsDirEvo"])

    # find all items in directory
    pathsH5 = directory.glob("*.h5")
    pathsPNG = directory.glob("*.png")

    # make set of items based on removing everything after _
    existsH5 = {str(path.name).split("_", 1)[0] for path in pathsH5}
    existsPNG = {str(path.name).split("_", 1)[0] for path in pathsPNG}

    toGenerate = existsH5 - existsPNG

    args = zip(
        toGenerate,
        itertools.repeat(sys.argv[1]),
        itertools.repeat(cache.settings["resultsDirEvo"]),
        itertools.repeat("%s_%s_EVO.png"),
    )

    list(map_function(plotExperiments, args))
    multiprocessing.get_logger().info("ending simulation graphs")


def graphMeta(cache, map_function):
    multiprocessing.get_logger().info("starting meta graphs")
    directory = Path(cache.settings["resultsDirMeta"])

    # find all items in directory
    pathsH5 = directory.glob("*.h5")
    pathsPNG = directory.glob("*.png")

    # make set of items based on removing everything after _
    existsH5 = {str(path.name).split("_", 1)[0] for path in pathsH5}
    existsPNG = {str(path.name).split("_", 1)[0] for path in pathsPNG}

    toGenerate = existsH5 - existsPNG

    args = zip(
        toGenerate,
        itertools.repeat(sys.argv[1]),
        itertools.repeat(cache.settings["resultsDirMeta"]),
        itertools.repeat("%s_%s_meta.png"),
    )

    list(map_function(plotExperiments, args))
    multiprocessing.get_logger().info("ending meta graphs")


def graph_simulation_unit(simulation, unit, graph):
    ncomp = int(simulation.root.input.model[unit].ncomp)
    isotherm = bytes(simulation.root.input.model[unit].adsorption_model)

    # This does not work correctly since we could be reading an outlet or tubing that does not have an isotherm
    # hasSalt = isotherm in saltIsotherms

    solution_times = simulation.root.output.solution.solution_times

    comps = []

    hasColumn = any("column" in i for i in simulation.root.output.solution[unit].keys())
    hasPort = any("port" in i for i in simulation.root.output.solution[unit].keys())

    if hasColumn:
        for i in range(ncomp):
            comps.append(
                simulation.root.output.solution[unit][
                    "solution_column_outlet_comp_%03d" % i
                ]
            )
    elif hasPort:
        for i in range(ncomp):
            comps.append(
                simulation.root.output.solution[unit][
                    "solution_outlet_port_000_comp_%03d" % i
                ]
            )
    else:
        for i in range(ncomp):
            comps.append(
                simulation.root.output.solution[unit]["solution_outlet_comp_%03d" % i]
            )

    # for now lets assume that if your first component is much higher concentration than the other components then the first component is salt
    max_comps = [max(comp) for comp in comps]

    hasSalt = any([max_comps[0] > (comp * 10) for comp in max_comps[1:]])

    if hasSalt:
        graph.set_title("Output %s" % unit)
        graph.plot(solution_times, comps[0], "b-", label="Salt")
        graph.set_xlabel("time (s)")

        # Make the y-axis label, ticks and tick labels match the line color.
        graph.set_ylabel("mM Salt", color="b")
        graph.tick_params("y", colors="b")

        axis2 = graph.twinx()
        for idx, comp in enumerate(comps[1:]):
            axis2.plot(
                solution_times,
                comp,
                "-",
                color=get_color(idx, len(comps) - 1, cm_plot),
                label="P%s" % idx,
            )
        axis2.set_ylabel("mM Protein", color="r")
        axis2.tick_params("y", colors="r")

        lines, labels = graph.get_legend_handles_labels()
        lines2, labels2 = axis2.get_legend_handles_labels()
        axis2.legend(lines + lines2, labels + labels2, loc=0)
    else:
        graph.set_title("Output %s" % unit)

        # colors = ['r', 'g', 'c', 'm', 'y', 'k']
        for idx, comp in enumerate(comps):
            graph.plot(
                solution_times,
                comp,
                "-",
                color=get_color(idx, len(comps), cm_plot),
                label="P%s" % idx,
            )
        graph.set_ylabel("mM Protein", color="r")
        graph.tick_params("y", colors="r")
        graph.set_xlabel("time (s)")

        lines, labels = graph.get_legend_handles_labels()
        graph.legend(lines, labels, loc=0)


def convert_frac(start, stop, values):
    x = []
    y = []
    for xl, xu, val in zip(start, stop, values):
        x.append(xl)
        y.append(val)
        x.append(xu)
        y.append(val)
    return x, y


def plotExperiments(args):
    save_name_base, json_path, directory, file_pattern = args
    if json_path != cache.json_path:
        cache.setup(json_path)

    target = cache.target
    settings = cache.settings
    for experiment in settings["experiments"]:
        experimentName = experiment["name"]

        dst = Path(directory, file_pattern % (save_name_base, experimentName))

        if dst.exists():
            # this item has already been plotted, this means that plots are occurring faster than generations are so kill this version
            sys.exit()

        units_used = cache.target[experimentName]["units_used"]

        numPlotsSeq = [len(units_used)]
        for feature in experiment["scores"]:
            settings = cache.scores[feature["type"]].get_settings(feature)
            numPlotsSeq.append(settings.graph_der + settings.graph + settings.graph_frac)

        numPlots = sum(numPlotsSeq)

        exp_time = target[experimentName]["time"]
        exp_value = target[experimentName]["value"]

        fig = figure.Figure(figsize=[24, 12 * numPlots])
        canvas = FigureCanvas(fig)

        simulation = Cadet()
        h5_path = Path(directory) / (
            file_pattern.replace("png", "h5") % (save_name_base, experimentName)
        )
        simulation.filename = h5_path.as_posix()
        simulation.load()

        results = {}
        results["simulation"] = simulation

        for idx, unit in enumerate(units_used):
            graph_simulation_unit(
                simulation, unit, fig.add_subplot(numPlots, 1, idx + 1)
            )

        graphIdx = idx + 2
        for idx, feature in enumerate(experiment["scores"]):
            featureName = feature["name"]
            featureType = feature["type"]

            feat = target[experimentName][featureName]

            selected = feat["selected"]
            exp_time = feat["time"][selected]
            exp_value = feat["value"][selected]

            sim_time, sim_value = get_times_values(
                simulation, target[experimentName][featureName]
            )

            settings = cache.scores[featureType].get_settings(feature)
            if settings.graph:
                graph = fig.add_subplot(
                    numPlots, 1, graphIdx
                )  # additional +1 added due to the overview plot
                graph.plot(sim_time, sim_value, "r--", label="Simulation")
                graph.plot(exp_time, exp_value, "g:", label="Experiment")
                graphIdx += 1

            if settings.graph_der:
                exp_spline = smoothing.smooth_data_derivative(
                    exp_time,
                    exp_value,
                    feat["critical_frequency"],
                    feat["smoothing_factor"],
                    feat["critical_frequency_der"],
                )
                sim_spline = smoothing.smooth_data_derivative(
                    sim_time,
                    sim_value,
                    feat["critical_frequency"],
                    feat["smoothing_factor"],
                    feat["critical_frequency_der"],
                )

                graph = fig.add_subplot(
                    numPlots, 1, graphIdx
                )  # additional +1 added due to the overview plot
                graph.plot(sim_time, sim_spline, "r--", label="Simulation")
                graph.plot(exp_time, exp_spline, "g:", label="Experiment")
                graphIdx += 1

            if settings.graph_frac:
                cache.scores[featureType].run(results, feat)

                graph_exp = results["graph_exp"]
                graph_sim = results["graph_sim"]
                graph_sim_offset = results.get("graph_sim_offset", {})

                findMax = 0
                max_comp = {}
                for idx, (key, value) in enumerate(graph_sim.items()):
                    (time_start, time_stop, values) = zip(*value)
                    max_value = max(values)
                    findMax = max(findMax, max_value)
                    max_comp[key] = max_value

                for idx, (key, value) in enumerate(graph_exp.items()):
                    (time_start, time_stop, values) = zip(*value)
                    max_value = max(values)
                    findMax = max(findMax, max_value)
                    max_comp[key] = max(max_comp[key], max_value)

                graph = fig.add_subplot(
                    numPlots, 1, graphIdx
                )  # additional +1 added due to the overview plot
                factors = []
                for idx, (key, value) in enumerate(graph_sim.items()):
                    (time_start, time_stop, values) = zip(*value)
                    values = numpy.array(values)
                    mult = 1.0
                    if max(values) < (0.2 * findMax):
                        mult = findMax / (2 * max_comp[key])
                        values = values * mult
                    factors.append(mult)

                    time_offset = graph_sim_offset.get(key, None)
                    if time_offset is not None:
                        if time_offset > 0:
                            time_str = "time early (s): %.3g" % abs(time_offset)
                        else:
                            time_str = "time late (s): %.3g" % abs(time_offset)
                    else:
                        time_str = ""

                    label = "Simulation Comp: %s Mult:%.2f %s" % (key, mult, time_str)

                    x, y = convert_frac(time_start, time_stop, values)

                    graph.plot(
                        x,
                        y,
                        "--",
                        color=get_color(idx, len(graph_sim), cm_plot),
                        label=label,
                    )

                for idx, (key, value) in enumerate(graph_exp.items()):
                    (time_start, time_stop, values) = zip(*value)
                    values = numpy.array(values)
                    mult = factors[idx]
                    values = values * mult

                    x, y = convert_frac(time_start, time_stop, values)

                    graph.plot(
                        x,
                        y,
                        ":",
                        color=get_color(idx, len(graph_sim), cm_plot),
                        label="Experiment Comp: %s Mult:%.2f" % (key, mult),
                    )
                graphIdx += 1
            graph.legend()
        fig.set_size_inches((24, 12 * numPlots))
        fig.savefig(str(dst))


def graphSpace(fullGeneration, cache, map_function, results):
    multiprocessing.get_logger().info("starting space graphs")
    output_2d = cache.settings["resultsDirSpace"] / "2d"
    output_3d = cache.settings["resultsDirSpace"] / "3d"
    output_prob = cache.settings["resultsDirSpace"] / "probability"

    output_2d.mkdir(parents=True, exist_ok=True)
    output_3d.mkdir(parents=True, exist_ok=True)
    output_prob.mkdir(parents=True, exist_ok=True)

    if results.root.is_extended_input:
        input = results.root.input_transform_extended
    else:
        input = results.root.input_transform

    if isinstance(results.root.probability, Dict):
        probability = None
    else:
        probability = numpy.squeeze(results.root.probability)
        probability_idx = numpy.max(probability) / probability < 1e5
        probability = probability[probability_idx]

    output = results.root.output
    output_meta = results.root.output_meta

    input_headers = cache.parameter_headers
    output_headers = cache.score_headers
    meta_headers = cache.meta_headers

    input_indexes = list(range(len(input_headers)))
    output_indexes = list(range(len(output_headers)))
    meta_indexes = list(range(len(meta_headers)))

    comp_two = list(itertools.combinations(input_indexes, 2))
    comp_one = list(itertools.combinations(input_indexes, 1))

    # 2d plots
    if fullGeneration >= 1:
        multiprocessing.get_logger().info("starting 2d space graphs")
        graphDistance(cache, map_function, results)

        seq = []
        for (x,), y in itertools.product(comp_one, output_indexes):
            seq.append(
                [
                    output_2d.as_posix(),
                    input_headers[x],
                    output_headers[y],
                    input[:, x],
                    output[:, y],
                ]
            )

        for (x,), y in itertools.product(comp_one, meta_indexes):
            seq.append(
                [
                    output_2d.as_posix(),
                    input_headers[x],
                    meta_headers[y],
                    input[:, x],
                    output_meta[:, y],
                ]
            )

        if probability is not None:
            for (x,) in comp_one:
                seq.append(
                    [
                        output_prob.as_posix(),
                        input_headers[x],
                        "probability",
                        input[probability_idx, x],
                        probability,
                    ]
                )

        list(map_function(plot_2d, seq))
        multiprocessing.get_logger().info("ending 2d space graphs")

    # 3d plots
    if fullGeneration >= 2:
        multiprocessing.get_logger().info("starting 3d space graphs")
        seq = []
        for (x, y), z in itertools.product(comp_two, output_indexes):
            seq.append(
                [
                    output_3d.as_posix(),
                    input_headers[x],
                    input_headers[y],
                    output_headers[z],
                    input[:, x],
                    input[:, y],
                    output[:, z],
                ]
            )

        for (x, y), z in itertools.product(comp_two, meta_indexes):
            seq.append(
                [
                    output_3d.as_posix(),
                    input_headers[x],
                    input_headers[y],
                    meta_headers[z],
                    input[:, x],
                    input[:, y],
                    output_meta[:, z],
                ]
            )
        list(map_function(plot_3d, seq))
        multiprocessing.get_logger().info("ending 3d space graphs")
    multiprocessing.get_logger().info("ending space graphs")


def plot_3d(arg):
    "This leaks memory and is run in a separate short-lived process, do not integrate into the matching main process"
    directory_path, header1, header2, header3, data1, data2, scores = arg
    directory = Path(directory_path)

    x_str = "%s"
    x = data1
    if max(x) / min(x) > 100:
        x_str = "log10(%s)"
        x = numpy.log10(x)

    y_str = "%s"
    y = data2
    if max(y) / min(y) > 100:
        y_str = "log10(%s)"
        y = numpy.log10(y)

    z_str = "%s"
    z = scores
    if max(z) / min(z) > 100:
        z_str = "log10(%s)"
        z = numpy.log10(z)

    fig = figure.Figure(figsize=[20, 20])
    canvas = FigureCanvas(fig)
    ax = fig.add_subplot(111, projection="3d")
    ax.scatter(x, y, z, c=z, cmap=my_cmap)
    ax.set_xlabel(x_str % header1)
    ax.set_ylabel(y_str % header2)
    ax.set_zlabel(z_str % header3)
    filename = "%s_%s_%s.png" % (header1, header2, header3)
    filename = filename.replace(":", "_").replace("/", "_")
    fig.savefig(str(directory / filename))


def plot_2d(arg):
    directory_path, header_x, scoreName, data, scores = arg

    multiprocessing.get_logger().info("2d graph %s %s start" % (header_x, scoreName))

    plot_2d_single(directory_path, header_x, scoreName, data, scores)

    multiprocessing.get_logger().info("2d graph %s %s stop" % (header_x, scoreName))


def plot_2d_single(directory_path, header_x, scoreName, data, scores):
    directory = Path(directory_path)

    fig = figure.Figure(figsize=[10, 10])
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)

    format = "%s"
    if numpy.max(data) / numpy.min(data[data > 0]) > 100.0:
        keep = data > 0
        data = numpy.log10(data[keep])
        scores = scores[keep]
        format = "log10(%s)"

    format_scores = "%s"
    if numpy.max(scores) / numpy.min(scores[scores > 0]) > 100.0:
        keep = scores > 0
        scores = numpy.log10(scores[keep])
        data = data[keep]
        format_scores = "log10(%s)"

    data_norm = (data - numpy.min(data)) / (numpy.max(data) - numpy.min(data))
    scores_norm = (scores - numpy.min(scores)) / (numpy.max(scores) - numpy.min(scores))
    scatter_data = numpy.array([data_norm, scores_norm]).T
    scatter_data_round = numpy.round(scatter_data, 2)
    scatter_data_round, idx, counts = numpy.unique(
        scatter_data_round, axis=0, return_index=True, return_counts=True
    )

    counts_norm = (counts - numpy.min(counts)) / (
        numpy.max(counts) - numpy.min(counts)
    ) * 100 + 10

    graph.scatter(data[idx], scores[idx], c=scores[idx], s=counts_norm, cmap=cmap)

    graph.set_xlabel(format % header_x)
    graph.set_ylabel(format_scores % scoreName)
    graph.set_xlim(min(data), max(data), auto=True)
    filename = "%s_%s.png" % (header_x, scoreName)
    filename = filename.replace(":", "_").replace("/", "_")

    fig.savefig(str(directory / filename))


def graphProgress(cache, map_function, df):
    multiprocessing.get_logger().info("starting progress graphs")
    output = cache.settings["resultsDirProgress"]

    x = [
        "Generation",
    ]
    y = ["Meta Front", "Meta Min", "Meta Product", "Meta Mean", "Meta SSE", "Meta RMSE"]

    temp = []
    for x, y in itertools.product(x, y):
        if x != y:
            temp.append((df[[x, y]], output))

    list(map_function(singleGraphProgress, temp))
    multiprocessing.get_logger().info("ending progress graphs")


def singleGraphProgress(arg):
    df, output = arg

    i, j = list(df)

    fig = figure.Figure(figsize=(10, 10))
    canvas = FigureCanvas(fig)

    graph = fig.add_subplot(1, 1, 1)

    x = df[i].to_numpy()
    y = df[j].to_numpy()

    title_str = "%s vs %s"
    y_label = "%s"
    if numpy.max(y) / numpy.min(y) > 100.0:
        y = numpy.log10(y)
        title_str = "%s vs log10(%s)"
        y_label = "log10(%s)"

    graph.plot(x, y)
    y_max = max(y)
    y_min = min(y)
    graph.set_ylim((1.1 * y_min, 1.1 * y_max))
    graph.set_title(title_str % (i, j))
    graph.set_xlabel(i)
    graph.set_ylabel(y_label % j)

    filename = "%s vs %s.png" % (i, j)
    file_path = output / filename
    fig.savefig(str(file_path))


def get_times_values(simulation, target, selected=None):

    times = simulation.root.output.solution.solution_times

    output_path = target["output_path"]

    if isinstance(output_path, list):
        values = numpy.sum([simulation[i] for i in output_path], 0)
    else:
        values = simulation[output_path]

    if selected is None:
        selected = target["selected"]

    return times[selected], values[selected] * target["factor"]


if __name__ == "__main__":
    map_function = util.getMapFunction()
    main(map_function)
    sys.exit()



import subprocess
import sys
from pathlib import Path
import pandas

def run_matching():
    results = []
    baseDir = Path(sys.argv[1]).resolve()
    for path in sorted(baseDir.rglob("results.xlsx")):
        print(path)
        df = pandas.read_excel(path)
        results.append(path.as_posix())
        results.append(df.to_string(index=False))

    with (baseDir / "results.txt").open("w", newline="") as outfile:
        outfile.write('\n\n'.join(results))

if __name__ == "__main__":
    run_matching()




import setuptools

from CADETMatch import version

with open("README.md", "r") as fh:
    long_description = fh.read()
    
setuptools.setup(
    name=version.__title__,
    version=version.__version__,
    author=version.__author__,
    author_email=version.__email__,
    description=version.__summary__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/modsim/CADET-Match",
    packages=setuptools.find_packages(),
    install_requires=[
          'joblib>=0.15.1',
          'addict>=2.2.1',
          'emcee>=3.0.2',
          'SAlib',
          'corner>=2.1.0',
          'pymoo>=0.5.0',
          'psutil>=5.7.0',
          'openpyxl>=3.0.3',
          'numpy>=1.18.5',
          'scipy>=1.5.0',
          'matplotlib>=3.2.1',
          'pandas>=1.0.5',
          'h5py>=2.10.0',
          'CADET-Python>=0.11',
          'seaborn>=0.10.1',
          'scikit-learn>=0.23.1',
          'importlib-metadata>=1.7.0',
          'jstyleson>=0.0.2',
          'attrs>=21.2.0',
          'filelock',
          'numba',
          'arviz>=0.11.2'
      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
) 


import multiprocessing
import subprocess
import sys
import time
from pathlib import Path

import filelock

import CADETMatch.util as util

processes = {}
times = {}


def get_lock(cache):
    resultDir = Path(cache.settings["resultsDirBase"])
    result_lock = resultDir / "result.lock"
    lock = filelock.FileLock(result_lock.as_posix())
    return lock


def run_sub(cache, key, line, file_name):
    sub = processes.get(key, None)

    if sub is None:
        multiprocessing.get_logger().info(
            "creating subprocess %s for %s", key, file_name
        )

        sub = subprocess.Popen(
            line,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        processes[key] = sub

    process_sub(key, file_name)


def process_sub(key, file_name):
    sub = processes.get(key, None)

    if sub is not None:
        finished = sub.poll() is not None
        if finished is not False:
            del processes[key]
            stdout, stderr = sub.communicate()
            multiprocessing.get_logger().info(
                "finished subprocess %s for %s", key, file_name
            )
            log_subprocess(file_name, stdout.decode("utf-8"), stderr.decode("utf-8"))


def wait_sub(key, file_name):
    sub = processes.get(key, None)
    if sub is not None:
        multiprocessing.get_logger().info(
            "waiting for subprocess %s for %s", key, file_name
        )
        stdout, stderr = sub.communicate()
        del processes[key]
        multiprocessing.get_logger().info(
            "finished subprocess %s for %s", key, file_name
        )
        log_subprocess(file_name, stdout.decode("utf-8"), stderr.decode("utf-8"))


def log_subprocess(name, stdout, stderr):
    for line in stdout.splitlines():
        multiprocessing.get_logger().info("%s stdout: %s", name, line)

    for line in stderr.splitlines():
        multiprocessing.get_logger().info("%s stderr: %s", name, line)

def graph_corner(cache):
    line = [
        sys.executable,
        (Path(__file__).parent / "generate_corner_graphs.py").as_posix(),
        str(cache.json_path),
        str(util.getCoreCounts()),
    ]
    run_sub(cache, "corner", line, "generate_corner_graphs.py")


def wait_corner():
    wait_sub("corner", "generate_corner_graphs.py")


def graph_autocorr(cache):
    line = [
        sys.executable,
        (Path(__file__).parent / "generate_autocorr_graphs.py").as_posix(),
        str(cache.json_path),
        str(util.getCoreCounts()),
    ]
    run_sub(cache, "autocorr", line, "generate_autocorr_graphs.py")


def wait_autocorr():
    wait_sub("autocorr", "generate_autocorr_graphs.py")


def graph_mixing(cache):
    line = [
        sys.executable,
        (Path(__file__).parent / "generate_mixing_graphs.py").as_posix(),
        str(cache.json_path),
        str(util.getCoreCounts()),
    ]
    run_sub(cache, "mixing", line, "generate_mixing_graphs.py")


def wait_mixing():
    wait_sub("mixing", "generate_mixing_graphs.py")


def graph_main(cache, graph_type):
    line = [
        sys.executable,
        (Path(__file__).parent / "generate_graphs.py").as_posix(),
        str(cache.json_path),
        graph_type,
        str(util.getCoreCounts()),
    ]
    run_sub(cache, "main", line, "generate_graphs.py")


def wait_main():
    wait_sub("main", "generate_graphs.py")

def wait_kde():
    wait_sub("graph_kde", "graph_kde.py")


def graph_kde(cache):
    line = [
        sys.executable,
        (Path(__file__).parent / "graph_kde.py").as_posix(),
        str(cache.json_path),
        str(util.getCoreCounts()),
    ]
    run_sub(cache, "graph_kde", line, "graph_kde.py")


def process_kde():
    process_sub("graph_kde", "graph_kde.py")


def graph_mle(cache):
    line = [sys.executable, 
        (Path(__file__).parent / "mle.py").as_posix(), 
        str(cache.json_path), 
        str(util.getCoreCounts())]
    run_sub(cache, "graph_mle", line, "mle.py")


def wait_mle():
    wait_sub("graph_mle", "mle.py")


def graph_prior(cache):
    line = [sys.executable, 
        (Path(__file__).parent / "gen_prior.py").as_posix(), 
        str(cache.json_path), 
        str(util.getCoreCounts())]
    run_sub(cache, "graph_prior", line, "gen_prior.py")


def wait_prior():
    wait_sub("graph_prior", "gen_prior.py")


def graph_tube(cache):
    line = [
        sys.executable,
        (Path(__file__).parent / "mcmc_plot_tube.py").as_posix(),
        str(cache.json_path),
        str(util.getCoreCounts()),
    ]
    run_sub(cache, "graph_tube", line, "mcmc_plot_tube.py")


def wait_tube():
    wait_sub("graph_tube", "mcmc_plot_tube.py")

def graph_process(cache, generation, last=False):
    lastGraphTime = times.get("lastGraphTime", time.time())
    lastMetaTime = times.get("lastMetaTime", time.time())

    if last:
        wait_main()

        graph_main(cache, str(cache.graphType))
        wait_main()
        lastGraphTime = time.time()

    elif (time.time() - lastGraphTime) > cache.graphGenerateTime:
        graph_main(cache, str(cache.graphType))
        lastGraphTime = time.time()
    else:
        if (time.time() - lastMetaTime) > cache.graphMetaTime:
            graph_main(cache, "0")
            lastMetaTime = time.time()

    times["lastGraphTime"] = lastGraphTime
    times["lastMetaTime"] = lastMetaTime


def graph_corner_process(cache, last=False, interval=3600):
    last_corner_time = times.get("last_corner_time", time.time())

    if last:
        process_kde()

        # there may be processes already running we need to wait for them and then run a last set on the final data set
        wait_corner()
        wait_autocorr()
        wait_mixing()

        graph_corner(cache)
        graph_autocorr(cache)
        graph_mixing(cache)

        wait_corner()
        wait_autocorr()
        wait_mixing()

        last_corner_time = time.time()
    elif (time.time() - last_corner_time) > interval:
        process_kde()
        graph_corner(cache)
        graph_autocorr(cache)
        graph_mixing(cache)

        last_corner_time = time.time()
    times["last_corner_time"] = last_corner_time


def mle_process(cache, last=False, interval=3600):
    last_mle_time = times.get("last_mle_time", time.time())

    if last:
        wait_mle()

        graph_mle(cache)
        wait_mle()

        last_mle_time = time.time()
    elif (time.time() - last_mle_time) > interval:
        graph_mle(cache)

        last_mle_time = time.time()

    times["last_mle_time"] = last_mle_time

def prior_process(cache, last=False, interval=3600):
    graph_prior(cache)
    wait_prior()


def tube_process(cache, last=False, interval=3600):
    last_tube_time = times.get("last_tube_time", time.time())

    if last:
        wait_tube()

        graph_tube(cache)
        wait_tube()

        last_tube_time = time.time()
    elif (time.time() - last_tube_time) > interval:
        graph_tube(cache)

        last_tube_time = time.time()

    times["last_tube_time"] = last_tube_time


import multiprocessing
import sys
import warnings
from pathlib import Path

import joblib
import matplotlib
import matplotlib.style as mplstyle
import numpy
import pandas
import scipy
from addict import Dict
from cadet import H5, Cadet
from sklearn import preprocessing
from sklearn.neighbors import KernelDensity

import CADETMatch.evo as evo
import CADETMatch.kde_util as kde_util
import CADETMatch.smoothing as smoothing
import CADETMatch.util as util
from CADETMatch.cache import cache

mplstyle.use("fast")

matplotlib.use("Agg")

import matplotlib.cm
import matplotlib.pyplot as plt
from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

cm_plot = matplotlib.cm.gist_rainbow

import itertools
import logging

import CADETMatch.loggerwriter as loggerwriter

import arviz as az

def get_color(idx, max_colors, cmap):
    return cmap(1.0 * float(idx) / max_colors)


saltIsotherms = {
    b"STERIC_MASS_ACTION",
    b"SELF_ASSOCIATION",
    b"MULTISTATE_STERIC_MASS_ACTION",
    b"SIMPLE_MULTISTATE_STERIC_MASS_ACTION",
    b"BI_STERIC_MASS_ACTION",
}

size = 20

plt.rc("font", size=size)  # controls default text sizes
plt.rc("axes", titlesize=size)  # fontsize of the axes title
plt.rc("axes", labelsize=size)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=size)  # fontsize of the tick labels
plt.rc("ytick", labelsize=size)  # fontsize of the tick labels
plt.rc("legend", fontsize=size)  # legend fontsize
plt.rc("figure", titlesize=size)  # fontsize of the figure title
plt.rc("figure", autolayout=True)

def flatten(chain):
    chain_shape = chain.shape
    flat_chain = chain.reshape(chain_shape[0] * chain_shape[1], chain_shape[2])
    return flat_chain

def get_mle(data_chain, probability, headers):
    flat_prob = probability.reshape(-1, 1)
    flat_chain = flatten(data_chain)
    best_idx = numpy.argmax(flat_prob)

    x = flat_chain[best_idx]

    multiprocessing.get_logger().info("mle found %s with probability %s", x, flat_prob[best_idx])

    return x


def addChain(axis, *args):
    temp = [arg for arg in args if len(arg)]
    if len(temp) > 1:
        return numpy.concatenate(temp, axis=axis)
    else:
        return numpy.array(temp[0])


def fitness(individual):
    return evo.fitness(individual, sys.argv[1])


def graph_simulations(simulations, simulation_labels, unit, graph):
    linestyles = ["-", "--", "-.", ":"]
    for idx_sim, (simulation, label_sim) in enumerate(
        zip(simulations, simulation_labels)
    ):

        comps = []

        ncomp = int(simulation.root.input.model.unit_001.ncomp)
        isotherm = bytes(simulation.root.input.model.unit_001.adsorption_model)

        hasSalt = isotherm in saltIsotherms

        solution_times = simulation.root.output.solution.solution_times

        hasColumn = any(
            "column" in i for i in simulation.root.output.solution[unit].keys()
        )
        hasPort = any("port" in i for i in simulation.root.output.solution[unit].keys())

        if hasColumn:
            for i in range(ncomp):
                comps.append(
                    simulation.root.output.solution[unit][
                        "solution_column_outlet_comp_%03d" % i
                    ]
                )
        elif hasPort:
            for i in range(ncomp):
                comps.append(
                    simulation.root.output.solution[unit][
                        "solution_outlet_port_000_comp_%03d" % i
                    ]
                )
        else:
            for i in range(ncomp):
                comps.append(
                    simulation.root.output.solution[unit][
                        "solution_outlet_comp_%03d" % i
                    ]
                )

        if hasSalt:
            graph.set_title("Output")
            graph.plot(solution_times, comps[0], "b-", label="Salt")
            graph.set_xlabel("time (s)")

            # Make the y-axis label, ticks and tick labels match the line color.
            graph.set_ylabel("mM Salt", color="b")
            graph.tick_params("y", colors="b")

            axis2 = graph.twinx()
            for idx, comp in enumerate(comps[1:]):
                axis2.plot(
                    solution_times,
                    comp,
                    linestyles[idx],
                    color=get_color(idx_sim, len(simulation_labels), cm_plot),
                    label="P%s %s" % (idx, label_sim),
                )
            axis2.set_ylabel("mM Protein", color="r")
            axis2.tick_params("y", colors="r")

            lines, labels = graph.get_legend_handles_labels()
            lines2, labels2 = axis2.get_legend_handles_labels()
            axis2.legend(lines + lines2, labels + labels2, loc=0)
        else:
            graph.set_title("Output")

            for idx, comp in enumerate(comps):
                graph.plot(
                    solution_times,
                    comp,
                    linestyles[idx],
                    color=get_color(idx_sim, len(simulation_labels), cm_plot),
                    label="P%s %s" % (idx, label_sim),
                )
            graph.set_ylabel("mM Protein", color="r")
            graph.tick_params("y", colors="r")
            graph.set_xlabel("time (s)")

            lines, labels = graph.get_legend_handles_labels()
            graph.legend(lines, labels, loc=0)


def plot_mle(simulations, cache, labels):
    mcmc_dir = Path(cache.settings["resultsDirMCMC"])
    target = cache.target
    settings = cache.settings
    for experiment in settings["experiments"]:
        experimentName = experiment["name"]

        file_name = "%s_stats.png" % experimentName
        dst = mcmc_dir / file_name

        units_used = cache.target[experimentName]["units_used"]

        numPlotsSeq = [len(units_used)]
        for feature in experiment["scores"]:
            settings = cache.scores[feature["type"]].get_settings(feature)
            numPlotsSeq.append(settings.graph_der + settings.graph + settings.graph_frac)

        numPlots = sum(numPlotsSeq)

        exp_time = target[experimentName]["time"]
        exp_value = target[experimentName]["valueFactor"]

        fig = figure.Figure(figsize=[15, 15 * numPlots])
        canvas = FigureCanvas(fig)

        for idx, unit in enumerate(units_used):
            graph_simulations(
                simulations[experimentName],
                labels,
                unit,
                fig.add_subplot(numPlots, 1, idx + 1),
            )

        graphIdx = 2
        for idx, feature in enumerate(experiment["scores"]):
            featureName = feature["name"]
            featureType = feature["type"]

            feat = target[experimentName][featureName]

            selected = feat["selected"]
            exp_time = feat["time"][selected]
            exp_value = feat["value"][selected]

            settings = cache.scores[featureType].get_settings(feature)
            if settings.graph:
                graph = fig.add_subplot(
                    numPlots, 1, graphIdx
                )  # additional +1 added due to the overview plot

                for idx, (sim, label) in enumerate(
                    zip(simulations[experimentName], labels)
                ):
                    sim_time, sim_value = util.get_times_values(
                        sim, target[experimentName][featureName]
                    )

                    if idx == 0:
                        linewidth = 2
                    else:
                        linewidth = 1

                    graph.plot(
                        sim_time,
                        sim_value,
                        "--",
                        label=label,
                        color=get_color(
                            idx, len(simulations[experimentName]) + 1, cm_plot
                        ),
                        linewidth=linewidth,
                    )

                graph.plot(
                    exp_time,
                    exp_value,
                    "-",
                    label="Experiment",
                    color=get_color(
                        len(simulations[experimentName]),
                        len(simulations[experimentName]) + 1,
                        cm_plot,
                    ),
                    linewidth=2,
                )
                graphIdx += 1

            if settings.graph_der:

                graph = fig.add_subplot(
                    numPlots, 1, graphIdx
                )  # additional +1 added due to the overview plot
                for idx, (sim, label) in enumerate(
                    zip(simulations[experimentName], labels)
                ):
                    sim_time, sim_value = util.get_times_values(
                        sim, target[experimentName][featureName]
                    )
                    sim_spline = smoothing.smooth_data_derivative(
                        sim_time,
                        sim_value,
                        feat["critical_frequency"],
                        feat["smoothing_factor"],
                        feat["critical_frequency_der"],
                    )

                    if idx == 0:
                        linewidth = 2
                    else:
                        linewidth = 1

                    graph.plot(
                        sim_time,
                        sim_spline,
                        "--",
                        label=label,
                        color=get_color(
                            idx, len(simulations[experimentName]) + 1, cm_plot
                        ),
                        linewidth=linewidth,
                    )

                exp_spline = smoothing.smooth_data_derivative(
                    exp_time,
                    exp_value,
                    feat["critical_frequency"],
                    feat["smoothing_factor"],
                    feat["critical_frequency_der"],
                )
                graph.plot(
                    exp_time,
                    exp_spline,
                    "-",
                    label="Experiment",
                    color=get_color(
                        len(simulations[experimentName]),
                        len(simulations[experimentName]) + 1,
                        cm_plot,
                    ),
                    linewidth=2,
                )
                graphIdx += 1

            graph.legend()

        fig.savefig(str(dst))


def process_mle(chain, probability, gen, cache):
    mcmc_dir = Path(cache.settings["resultsDirMCMC"])

    mcmc_csv = mcmc_dir / "prob.csv"
    mle_h5 = mcmc_dir / "mle.h5"

    h5 = H5()
    h5.filename = mle_h5.as_posix()
    if mle_h5.exists():
        h5.load(lock=True)

        if 0: #h5.root.generations[-1] == gen:
            multiprocessing.get_logger().info(
                "new information is not yet available and mle will quit"
            )
            return

    mle_x = get_mle(chain, probability, cache.parameter_headers_actual)

    multiprocessing.get_logger().info("mle_x: %s", mle_x)

    mle_ind = util.convert_individual_inputorder(mle_x, cache)

    multiprocessing.get_logger().info("mle_ind: %s", mle_ind)

    temp = [
        mle_x,
    ]

    multiprocessing.get_logger().info("chain shape: %s", chain.shape)

    # run simulations for 5% 50% 95% and MLE vs experimental data
    percentile_splits = [5, 10, 50, 90, 95]
    percentile = numpy.percentile(flatten(chain), percentile_splits, 0)

    multiprocessing.get_logger().info("percentile: %s %s", percentile.shape, percentile)

    for row in percentile:
        temp.append(list(row))

    cadetValues = [util.convert_individual_inputorder(i, cache) for i in temp]
    cadetValues = numpy.array(cadetValues)

    multiprocessing.get_logger().info(
        "cadetValues: %s %s", cadetValues.shape, cadetValues
    )

    map_function = util.getMapFunction()

    fitnesses = list(map_function(fitness, temp))

    simulations = {}
    for scores, csv_record, meta_score, results, individual in fitnesses:
        for name, value in results.items():
            sims = simulations.get(name, [])
            sims.append(value["simulation"])

            simulations[name] = sims

    multiprocessing.get_logger().info(
        "type %s  value %s", type(cadetValues), cadetValues
    )

    pd = pandas.DataFrame(cadetValues, columns=cache.parameter_headers_actual)
    labels = ["MLE", "5", "10", "50", "90", "95"]
    pd.insert(0, "name", labels)
    pd.to_csv(mcmc_csv, index=False)

    h5.root.stat_labels = cache.parameter_headers_actual
    h5.root.percentile_splits = percentile_splits

    mle_x = numpy.array(mle_x)

    h5.root.mles = addChain(1, h5.root.mles, mle_x[:, numpy.newaxis])
    h5.root.stats = addChain(2, h5.root.stats, percentile[:, :, numpy.newaxis])
    h5.root.generations = addChain(
        0,
        h5.root.generations,
        [
            gen,
        ],
    )
    h5.root.stat_MLE = mle_ind

    h5.save(lock=True)

    plot_mle(simulations, cache, labels)


def main():
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings["resultsDirLog"], "mle.log")
    cache.setup(sys.argv[1])

    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    mcmc_h5 = mcmcDir / "mcmc.h5"

    mcmc_store = H5()
    mcmc_store.filename = mcmc_h5.as_posix()
    mcmc_store.load(paths=["/full_chain", "/mcmc_acceptance", "/run_probability"], lock=True)

    process_mle(mcmc_store.root.full_chain, 
                mcmc_store.root.run_probability,
                len(mcmc_store.root.mcmc_acceptance), cache)


if __name__ == "__main__":
    main()



# This module simulates noise (systemic and random) to use for kernel density estimation based on experimental data

# Currently the randomness is

# Pump flow rate
# Pump delay
# Base noise
# Signal noise

import copy
import itertools
import multiprocessing
from pathlib import Path

import numpy
from cadet import H5, Cadet
from sklearn import preprocessing
from sklearn.neighbors import KernelDensity

import CADETMatch.kde_util as kde_util
import CADETMatch.score as score

import sys
import warnings

import joblib
import SALib.sample.sobol_sequence
import scipy.optimize
import scipy.stats

import CADETMatch.cache as cache
import CADETMatch.synthetic_error as synthetic_error
import CADETMatch.util as util
import CADETMatch.sub as sub


def mirror(data, meta_mask):
    data_min = numpy.min(data, 0)

    mirror_index = meta_mask
    keep_index = numpy.logical_not(meta_mask)

    data_mask = numpy.ma.masked_equal(data_min, 0.0, copy=False)
    min_value = data_mask.min(axis=0)

    data_mirror = numpy.copy(data)

    data_mirror[:, mirror_index] = -data_mirror[:, mirror_index]
    full_data = numpy.vstack([data_mirror, data])

    return full_data


def setupKDE(cache):
    scores = generate_synthetic_error(cache)

    mcmcDir = Path(cache.settings["resultsDirMCMC"])

    scores_mirror = mirror(scores, cache.meta_mask)

    scaler = getScaler(scores_mirror)

    scores_scaler = scaler.transform(scores_mirror)

    kde = KernelDensity(kernel="gaussian")

    kde, bandwidth, store = kde_util.get_bandwidth(kde, scores_scaler)

    kde.fit(scores_scaler)

    probability = kde.score_samples(scores_scaler)

    joblib.dump(scaler, mcmcDir / "kde_scaler.joblib")

    joblib.dump(kde, mcmcDir / "kde_score.joblib")

    h5_data = H5()
    h5_data.filename = (mcmcDir / "kde_settings.h5").as_posix()
    h5_data.root.bandwidth = bandwidth
    h5_data.root.store = numpy.array(store)
    h5_data.root.scores = scores
    h5_data.root.scores_mirror = scores_mirror
    h5_data.root.scores_mirror_scaled = scores_scaler
    h5_data.root.probability = probability

    h5_data.save(lock=True)

    sub.graph_kde(cache)

    return kde, scaler


def getScaler(data):
    scaler = preprocessing.RobustScaler().fit(data)
    return scaler


def getKDE(cache):
    mcmcDir = Path(cache.settings["resultsDirMCMC"])

    kde = joblib.load(mcmcDir / "kde_score.joblib")

    scaler = joblib.load(mcmcDir / "kde_scaler.joblib")

    return kde, scaler


def getKDEPrevious(cache):
    if "mcmc_h5" in cache.settings:
        mcmc_h5 = Path(cache.settings["mcmc_h5"])
        mcmcDir = mcmc_h5.parent

        if mcmcDir.exists():
            kde = joblib.load(mcmcDir / "kde_prior.joblib")

            scaler = joblib.load(mcmcDir / "kde_prior_scaler.joblib")

            return kde, scaler
    return None, None


def generate_data(cache):
    scores = generate_synthetic_error(cache)

    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    save_scores = mcmcDir / "scores_used.npy"

    numpy.save(save_scores, scores)

    scores_mirror = mirror(scores)

    scaler = getScaler(scores_mirror)

    scores_scaler = scaler.transform(scores_mirror)

    bandwidth, store = get_bandwidth(scores_scaler)

    return scores, bandwidth


def synthetic_error_simulation(x):
    json_path, error = x
    if json_path != cache.cache.json_path:
        cache.cache.setup_dir(json_path)
        util.setupLog(cache.cache.settings["resultsDirLog"], "main.log")
        cache.cache.setup(json_path)

    scores = []
    outputs = {}
    simulations = {}
    experiment_failed = False

    delays_store = []
    flows_store = []
    load_store = []
    uv_store = {}
    uv_store_norm = {}

    for experiment in cache.cache.settings["errorModel"]:
        delay_settings = experiment["delay"]
        flow_settings = experiment["flow"]
        load_settings = experiment["load"]
        experimental_csv = experiment["experimental_csv"]
        uv_noise = experiment.get("uv_noise", None)
        uv_noise_norm = experiment.get("uv_noise_norm", None)
        units = experiment["units"]
        name = experiment["name"]

        data = numpy.loadtxt(experimental_csv, delimiter=",")
        times = data[:, 0]

        if "file_path" in experiment:
            template_path = Path(experiment["file_path"])
        else:
            resultsDir = cache.cache.settings["resultsDir"]
            if "resultsDirOriginal" in cache.cache.settings:
                resultsDir = Path(cache.cache.settings["resultsDirOriginal"])

            template_path = resultsDir / "misc" / ("template_%s_base.h5" % name)

        temp = Cadet()
        temp.filename = template_path.as_posix()
        temp.load()

        util.setupSimulation(temp, times, name, cache.cache)

        nsec = temp.root.input.solver.sections.nsec

        for unit in units:
            unit_name = "unit_%03d" % unit
            for comp in range(temp.root.input.model[unit_name].ncomp):
                comp_name = "solution_outlet_comp_%03d" % comp

                if uv_noise is None:
                    error_uv = numpy.zeros(
                        len(temp.root.input.solver.user_solution_times)
                    )
                else:
                    error_uv = numpy.random.normal(
                        uv_noise[0],
                        uv_noise[1],
                        len(temp.root.input.solver.user_solution_times),
                    )
                uv_store["%s_%s_%s" % (name, unit, comp)] = error_uv

                if uv_noise_norm is None:
                    error_uv = numpy.ones(
                        len(temp.root.input.solver.user_solution_times)
                    )
                else:
                    error_uv = numpy.random.normal(
                        uv_noise_norm[0],
                        uv_noise_norm[1],
                        len(temp.root.input.solver.user_solution_times),
                    )
                uv_store_norm["%s_%s_%s" % (name, unit, comp)] = error_uv

        def post_function(simulation):
            # baseline drift need to redo this
            # error_slope = numpy.random.normal(error_slope_settings[0], error_slope_settings[1], 1)[0]

            for unit in units:
                unit_name = "unit_%03d" % unit
                for comp in range(simulation.root.input.model[unit_name].ncomp):
                    comp_name = "solution_outlet_comp_%03d" % comp
                    error_uv = uv_store["%s_%s_%s" % (name, unit, comp)]
                    error_norm = uv_store_norm["%s_%s_%s" % (name, unit, comp)]
                    simulation.root.output.solution[unit_name][comp_name] = (
                        simulation.root.output.solution[unit_name][comp_name]
                        * error_norm
                        + error_uv
                    )

        error_delay = Cadet(temp.root)

        # delays = numpy.random.uniform(delay_settings[0], delay_settings[1], nsec)
        delays = error[name]["pump_delays"]
        delays_store.extend(delays)

        synthetic_error.pump_delay(error_delay, delays)

        # flow = numpy.random.normal(flow_settings[0], flow_settings[1], error_delay.root.input.solver.sections.nsec)
        flow = error[name]["flow_rates"]
        flows_store.extend(flow)

        synthetic_error.error_flow(error_delay, flow)

        # load = numpy.random.normal(load_settings[0], load_settings[1], error_delay.root.input.solver.sections.nsec)
        load = error[name]["loading_concentrations"]
        load_store.extend(load)

        synthetic_error.error_load(error_delay, load)

        exp_info = None
        for exp in cache.cache.settings["experiments"]:
            if exp["name"] == name:
                exp_info = exp
                break

        result = util.runExperiment(
            None,
            exp_info,
            cache.cache.settings,
            cache.cache.target,
            error_delay,
            error_delay.root.timeout,
            cache.cache,
            post_function=post_function,
        )

        if result is not None:
            scores.extend(result["scores"])

            simulations[name] = result["simulation"]

            for unit in units:
                unit_name = "unit_%03d" % unit
                for comp in range(
                    result["simulation"].root.input.model[unit_name].ncomp
                ):
                    outputs[
                        "%s_unit_%03d_comp_%03d" % (name, int(unit), comp)
                    ] = result["simulation"].root.output.solution[
                        "unit_%03d" % int(unit)
                    ][
                        "solution_outlet_comp_%03d" % comp
                    ]
        else:
            experiment_failed = True

    if experiment_failed:
        return None, None, None, None

    errors = {}
    errors["delays"] = delays_store
    errors["flows"] = flows_store
    errors["load"] = load_store
    errors["uv_store"] = uv_store

    return scores, simulations, outputs, errors


def get_section_counts(cache):
    sections = []

    for experiment in cache.cache.settings["errorModel"]:
        name = experiment["name"]

        if "file_path" in experiment:
            template_path = Path(experiment["file_path"])
        else:
            resultsDir = cache.cache.settings["resultsDir"]
            if "resultsDirOriginal" in cache.cache.settings:
                resultsDir = Path(cache.cache.settings["resultsDirOriginal"])

            template_path = resultsDir / "misc" / ("template_%s_base.h5" % name)

        temp = Cadet()
        temp.filename = template_path.as_posix()
        temp.load()

        nsec = temp.root.input.solver.sections.nsec

        sections.append(nsec)

    return sections


def generate_error_sequence(cache):
    errors = {}
    count = cache.settings["errorModelCount"]
    for experiment in cache.settings["errorModel"]:
        delay_settings = experiment["delay"]
        flow_settings = experiment["flow"]
        load_settings = experiment["load"]
        name = experiment["name"]

        if "file_path" in experiment:
            template_path = Path(experiment["file_path"])
        else:
            resultsDir = cache.cache.settings["resultsDir"]
            if "resultsDirOriginal" in cache.settings:
                resultsDir = Path(cache.settings["resultsDirOriginal"])

            template_path = resultsDir / "misc" / ("template_%s_base.h5" % name)

        temp = Cadet()
        temp.filename = template_path.as_posix()
        temp.load()

        nsec = temp.root.input.solver.sections.nsec

        error = {}

        # total space is 1 delay per section 1 flow rate per new section and 1 load per new section
        delay_size = nsec
        flow_size = 2 * nsec
        load_size = 2 * nsec
        total_size = delay_size + flow_size + load_size

        errors_exp = SALib.sample.sobol_sequence.sample(count, total_size)

        # set delays
        block = errors_exp[:, :delay_size]
        if delay_settings[0] != delay_settings[1]:
            block = scipy.stats.uniform.ppf(
                block, delay_settings[0], delay_settings[1] - delay_settings[0]
            )
        else:
            block = numpy.ones(block.shape) * delay_settings[0]

        error["pump_delays"] = block

        # set flow range changes
        lb = delay_size
        ub = delay_size + flow_size

        # grab the block to work with easier
        block = errors_exp[:, lb:ub]

        # sobol sequence is [0, 1) and 0 needs to be removed and set to the same distance as the closest distance to 1
        min_value = 1 - numpy.max(block)

        block[block < min_value] = min_value

        if flow_settings[1] != 0:
            block = scipy.stats.norm.ppf(block, flow_settings[0], flow_settings[1])
        else:
            block = numpy.ones(block.shape) * flow_settings[0]

        error["flow_rates"] = block

        # set loading changes
        lb = ub
        ub = lb + load_size

        # grab the block to work with easier
        block = errors_exp[:, lb:ub]

        # sobol sequence is [0, 1) and 0 needs to be removed and set to the same distance as the closest distance to 1
        min_value = 1 - numpy.max(block)

        block[block < min_value] = min_value

        if load_settings[1] != 0:
            block = scipy.stats.norm.ppf(block, load_settings[0], load_settings[1])
        else:
            block = numpy.ones(block.shape) * load_settings[0]

        error["loading_concentrations"] = block

        errors[name] = error

    return errors


def split_errors(errors_all):
    errors = []
    for name, experiment in errors_all.items():
        length = len(experiment["flow_rates"])
        break

    temp = {}
    for key in errors_all:
        temp[key] = {}

    for i in range(length):
        errors.append({key: {} for key in errors_all})

    for name, experiment in errors_all.items():
        for error_name, error_value in experiment.items():
            for idx in range(length):
                errors[idx][name][error_name] = error_value[idx, :]

    return errors


def keep_data(data, lb=0.5, ub=99.5):
    cutoffs_lb, cutoffs_ub = numpy.percentile(data, [lb, ub], axis=0)
    remove = numpy.zeros(data.shape)

    for i in range(data.shape[1] - 2):
        remove[:, i] = data[:, i] > cutoffs_lb[i]
    remove[:, -2] = (data[:, -2] > cutoffs_lb[-2]) & (data[:, -2] < cutoffs_ub[-2])
    remove[:, -1] = (data[:, -1] > cutoffs_lb[-1]) & (data[:, -1] < cutoffs_ub[-1])

    remove = numpy.all(remove, axis=1)

    return remove


def generate_synthetic_error(cache):
    count_settings = int(cache.settings["errorModelCount"])
    if "errorModel" in cache.settings:
        scores_all = []
        times = {}
        outputs_all = {}

        uv_store_all = {}

        errors_all = generate_error_sequence(cache)

        errors_split = split_errors(errors_all)
        json_path = itertools.repeat(cache.json_path)

        indexes = []

        for idx, (scores, simulations, outputs, errors) in enumerate(
            cache.map_function(synthetic_error_simulation, zip(json_path, errors_split))
        ):
            if scores and simulations and outputs:
                indexes.append(idx)

                for key, value in errors["uv_store"].items():
                    uv = uv_store_all.get(key, [])
                    uv.append(value)
                    uv_store_all[key] = uv

                scores_all.append(scores)

                for key, value in outputs.items():
                    temp = outputs_all.get(key, [])
                    temp.append(value)
                    outputs_all[key] = temp

                for key, sim in simulations.items():
                    if key not in times:
                        times[key] = sim.root.output.solution.solution_times

        indexes = numpy.array(indexes)

        for name, experiment in errors_all.items():
            for error_name, error_value in experiment.items():
                errors_all[name][error_name] = error_value[indexes, :]

        scores = numpy.array(scores_all)

        dir_base = cache.settings.get("resultsDirBase")
        file = dir_base / "kde_data.h5"

        kde_data = H5()
        kde_data.filename = file.as_posix()

        kde_data.root.scores = scores

        for output_name, output in outputs_all.items():
            kde_data.root[output_name] = numpy.array(output)

        for time_name, time in times.items():
            kde_data.root["%s_time" % time_name] = time

        for name, experiment in errors_all.items():
            for error_name, error_value in experiment.items():
                kde_data.root.errors[name][error_name] = error_value

        for key, value in uv_store_all.items():
            kde_data.root.uv_store[key] = numpy.array(value)

        kde_data.save(lock=True)

        return scores

    return None


def convert_to_array(seq):
    "this converts a sequence of arrays to a single numpy array and pads uneven rows with 0"
    max_len = numpy.max([len(a) for a in seq])
    return numpy.asarray(
        [
            numpy.pad(a, (0, max_len - len(a)), "constant", constant_values=0)
            for a in seq
        ]
    )



import multiprocessing
import warnings

import numpy
import numpy.linalg
import scipy
import scipy.signal
from cadet import H5

import CADETMatch.util as util

from pymoo.core.problem import ElementwiseProblem
from pymoo.factory import get_algorithm, get_reference_directions
from pymoo.optimize import minimize

butter_order = 3

class TargetProblem(ElementwiseProblem):

    def __init__(self, lb, ub, sse_target, func, values, fs):
        super().__init__(n_var=1, n_obj=1, n_constr=0, xl=lb, xu=ub)
        self.sse_target = sse_target
        self.func = func
        self.values = values
        self.fs = fs        

    def _evaluate(self, crit_fs, out, *args, **kwargs):
        crit_fs = 10**crit_fs
        try:
            sos = self.func(crit_fs, self.fs)
            low_passed = scipy.signal.sosfiltfilt(sos, self.values)
            sse = numpy.sum((low_passed - self.values) ** 2)

            error = (sse - self.sse_target)**2
        except ValueError:
            error = numpy.inf
        out["F"] = error

class MaxDistance(ElementwiseProblem):

    def __init__(self, lb, ub, func, fs, values, x_min, y_min, p1, p2, factor):
        super().__init__(n_var=1, n_obj=1, n_constr=0, xl=lb, xu=ub)
        self.func = func
        self.fs = fs   
        self.values = values
        self.x_min = x_min
        self.y_min = y_min
        self.p1 = p1
        self.p2 = p2
        self.factor = factor

    def _evaluate(self, crit_fs, out, *args, **kwargs):
        crit_fs = 10.0 ** crit_fs[0]
        try:
            sos = self.func(crit_fs, self.fs)
        except ValueError:
            return 1e6
        low_passed = scipy.signal.sosfiltfilt(sos, self.values)

        sse = numpy.sum((low_passed - self.values) ** 2)

        pT = numpy.array([crit_fs - self.x_min, numpy.log(sse) - self.y_min]).T / self.factor

        d = numpy.cross(self.p2 - self.p1, self.p1 - pT) / numpy.linalg.norm(self.p2 - self.p1)
        out["F"] = -d

def get_p(x, y):
    x = numpy.array(x)
    y = numpy.array(y)

    sort_idx = numpy.argsort(x)

    x = x[sort_idx]
    y = y[sort_idx]

    y_min = y - min(y)
    x_min = x - min(x)

    p3 = numpy.array([x_min, y_min]).T
    factor = numpy.max(p3, 0)
    p3 = p3 / factor
    p1 = p3[0, :]
    p2 = p3[-1, :]

    return x, min(x), y, min(y), p1, p2, p3, factor

def signal_bessel(crit_fs, fs):
    return scipy.signal.bessel(butter_order, crit_fs, btype="lowpass", analog=False, fs=fs, output="sos", norm="delay")

def signal_butter(crit_fs, fs):
    return scipy.signal.butter(butter_order, crit_fs, btype="lowpass", analog=False, fs=fs, output="sos")

def refine_signal(func, times, values, x, y, fs, start):
    x, x_min, y, y_min, p1, p2, p3, factor = get_p(x, y)

    lb = numpy.log10(x[0])
    ub = numpy.log10(x[-1])

    problem = MaxDistance(lb, ub, func, fs, values, x_min, y_min, p1, p2, factor)

    algorithm = get_algorithm('pattern-search', n_sample_points=50, eps=1e-13)

    res = minimize(problem,
               algorithm,
               verbose=False,
               seed=1)

    crit_fs = 10**res.X[0]

    return crit_fs


def find_L(x, y):
    # find the largest value greater than 0, otherwise return none to just turn off butter filter
    x, x_min, y, y_min, p1, p2, p3, factor = get_p(x, y)

    d = numpy.cross(p2 - p1, p1 - p3) / numpy.linalg.norm(p2 - p1)

    max_idx = numpy.argmax(d)
    max_d = d[max_idx]
    l_x = x[max_idx]
    l_y = y[max_idx]

    if max_d <= 0:
        return None, None

    return l_x, l_y


def find_signal(func, times, values, sse_target):
    filters = []
    sse = []

    fs = 1.0 / (times[1] - times[0])

    ub = fs / 2.0

    for i in numpy.logspace(-8, numpy.log10(ub), 50):
        try:
            sos = func(i, fs)
            low_passed = scipy.signal.sosfiltfilt(sos, values)

            filters.append(i)
            sse.append(numpy.sum((low_passed - values) ** 2))
        except (ValueError, numpy.linalg.LinAlgError):
            continue

    crit_fs_max = find_max_signal(func, times, values, sse_target, filters, sse)

    L_x, L_y = find_L(filters, numpy.log(sse))

    if L_x is not None:
        L_x = refine_signal(func, times, values, filters, numpy.log(sse), fs, L_x)

    if L_x  is not None and crit_fs_max < L_x:
        L_x = crit_fs_max

    return L_x

def find_max_signal(func, times, values, sse_target, filters, sse):
    fs = 1.0 / (times[1] - times[0])

    filters = numpy.log10(filters)
    problem = TargetProblem(filters[0], filters[-1], sse_target, func, values, fs)

    algorithm = get_algorithm('pattern-search', n_sample_points=50, eps=1e-13)

    res = minimize(problem,
               algorithm,
               verbose=False,
               seed=1)

    crit_fs = 10**res.X[0]

    return crit_fs


def smoothing_filter_signal(func, times, values, crit_fs):
    if crit_fs is None:
        return values
    fs = 1.0 / (times[1] - times[0])
    sos = func(crit_fs, fs)
    low_passed = scipy.signal.sosfiltfilt(sos, values)
    return low_passed


def load_data(name, cache):
    crit_fs = None
    crit_fs_der = None
    s = None
    s_knots = 0

    # quick abort
    if name is None or cache is None:
        return s, crit_fs, crit_fs_der

    factor_file = cache.settings["resultsDirMisc"] / "find_smoothing_factor.h5"

    data = H5()
    data.filename = factor_file.as_posix()

    if factor_file.exists():
        data.load(lock=True)

    if name in data.root:
        s = float(data.root[name].s)

        crit_fs = data.root[name].crit_fs
        if crit_fs == -1.0:
            crit_fs = None

        crit_fs_der = data.root[name].crit_fs_der
        if crit_fs_der == -1.0:
            crit_fs_der = None

        s_knots = int(data.root[name].s_knots)
    else:
        return s, crit_fs, crit_fs_der

    crit_fs_message = "critical frequency disable"
    if crit_fs is not None:
        crit_fs_message = "critical frequency %.3e" % crit_fs

    crit_fs_der_message = "critical frequency der disable"
    if crit_fs_der is not None:
        crit_fs_der_message = "critical frequency der %.3e" % crit_fs_der

    multiprocessing.get_logger().info(
        "smoothing_factor %s  %.3e  %s  %s knots %d",
        name,
        s,
        crit_fs_message,
        crit_fs_der_message,
        s_knots,
    )

    return s, crit_fs, crit_fs_der


def find_smoothing_factors(times, values, name, cache, rmse_target=1e-4):
    times, values = resample(times, values)
    min = 1e-3

    sse_target = (rmse_target**2.0)*len(values)

    s, crit_fs, crit_fs_der = load_data(name, cache)

    if s is not None:
        return s, crit_fs, crit_fs_der

    # normalize the data
    values = values * 1.0 / max(values)

    crit_fs = find_signal(signal_bessel, times, values, sse_target)

    if crit_fs is None:
        multiprocessing.get_logger().info(
            "%s butter filter disabled, no viable L point found", name
        )

    values_filter = smoothing_filter_signal(signal_bessel, times, values, crit_fs)

    s = sse_target
 
    spline, factor = create_spline(times, values, crit_fs, s)

    # run a quick butter pass to remove high frequency noise in the derivative (needed for some experimental data)
    values_filter = spline.derivative()(times) / factor
    factor = 1.0/numpy.max(values_filter)
    values_filter = values_filter * factor
    crit_fs_der = find_signal(signal_bessel, times, values_filter, sse_target)

    s_knots = 0
    knots = spline.get_knots()
    all_s = [s, s]

    record_smoothing(s, s_knots, crit_fs, crit_fs_der, knots, all_s, name, cache)

    return s, crit_fs, crit_fs_der


def record_smoothing(
    s, s_knots, crit_fs, crit_fs_der, knots, all_s, name=None, cache=None
):
    if name is None or cache is None:
        return
    factor_file = cache.settings["resultsDirMisc"] / "find_smoothing_factor.h5"

    data = H5()
    data.filename = factor_file.as_posix()

    if factor_file.exists():
        data.load(lock=True)

    if name not in data.root:
        data.root[name].knots = knots
        data.root[name].all_s = all_s
        data.root[name].s = float(s)
        data.root[name].s_knots = int(s_knots)
        if crit_fs is None:
            data.root[name].crit_fs = -1.0
        else:
            data.root[name].crit_fs = float(crit_fs)
        if crit_fs_der is None:
            data.root[name].crit_fs_der = -1.0
        else:
            data.root[name].crit_fs_der = float(crit_fs)
        data.save(lock=True)

    crit_fs_message = "critical frequency disable"
    if crit_fs is not None:
        crit_fs_message = "critical frequency %.3e" % crit_fs

    crit_fs_der_message = "critical frequency der disable"
    if crit_fs_der is not None:
        crit_fs_der_message = "critical frequency der %.3e" % crit_fs_der

    multiprocessing.get_logger().info(
        "smoothing_factor %s  %.3e  %s  %s knots %d",
        name,
        s,
        crit_fs_message,
        crit_fs_der_message,
        s_knots,
    )


def create_spline(times, values, crit_fs, s):
    times, values = resample(times, values)
    factor = 1.0 / numpy.max(values)
    values = values * factor
    values_filter = smoothing_filter_signal(signal_bessel, times, values, crit_fs)

    return (
        scipy.interpolate.UnivariateSpline(times, values_filter, s=s, k=5, ext=3),
        factor,
    )


def smooth_data(times, values, crit_fs, s):
    spline, factor = create_spline(times, values, crit_fs, s)

    return spline(times) / factor


def smooth_data_derivative(times, values, crit_fs, s, crit_fs_der, smooth=True):
    times_resample, values_resample = resample(times, values)
    spline, factor = create_spline(times_resample, values_resample, crit_fs, s)

    if smooth:
        values_filter_der = spline.derivative()(times_resample) / factor

        factor_der = 1.0/numpy.max(values_filter_der)
        values_filter_der = values_filter_der * factor_der
        values_filter_der = butter(times_resample, values_filter_der, crit_fs_der)
        values_filter_der = values_filter_der / factor_der
        spline_der = scipy.interpolate.InterpolatedUnivariateSpline(
            times_resample, values_filter_der, k=5, ext=3
        )
        values_filter_der = spline_der(times)
    else:
        values_filter_der = spline.derivative()(times) / factor
    return values_filter_der


def full_smooth(times, values, crit_fs, s, crit_fs_der, smooth=True):
    # return smooth data derivative of data
    times_resample, values_resample = resample(times, values)
    spline, factor = create_spline(times_resample, values_resample, crit_fs, s)

    values_filter = spline(times) / factor

    # run a quick butter pass to remove high frequency noise in the derivative (needed for some experimental data)

    if smooth:
        values_filter_der = spline.derivative()(times_resample) / factor

        factor_der = 1.0/numpy.max(values_filter_der)
        values_filter_der = values_filter_der * factor_der
        values_filter_der = butter(times_resample, values_filter_der, crit_fs_der)
        values_filter_der = values_filter_der / factor_der
        spline_der = scipy.interpolate.InterpolatedUnivariateSpline(
            times_resample, values_filter_der, k=5, ext=3
        )
        values_filter_der = spline_der(times)
    else:
        values_filter_der = spline.derivative()(times) / factor
    return values_filter, values_filter_der


def butter(times, values, crit_fs_der):
    max_value = numpy.max(values)
    values = values / max_value

    values_filter = smoothing_filter_signal(signal_bessel, times, values, crit_fs_der) * max_value

    return values_filter


def resample(times, values, max_samples=5000):
    if len(times) > max_samples:
        times_resample = numpy.linspace(times[0], times[-1], max_samples)
        spline_resample = scipy.interpolate.InterpolatedUnivariateSpline(times, values, k=5, ext=3)
        values_resample = spline_resample(times_resample)

        return times_resample, values_resample
    else: 
        diff_times = times[1:] - times[:-1]
        max_time = numpy.max(diff_times)
        min_time = numpy.min(diff_times)
        per = (max_time - min_time) / min_time

        if per > 0.01:
            # time step is not consistent, resample the time steps to a uniform grid based on the smallest time step size seen
            #but not more samples than max_samples
            times_resample = numpy.arange(times[0], times[-1], min_time)

            if len(times_resample) > max_samples:
               times_resample = numpy.linspace(times[0], times[-1], max_samples) 

            times_resample[-1] = times[-1]
            spline_resample = scipy.interpolate.InterpolatedUnivariateSpline(times, values, k=5, ext=3)
            values_resample = spline_resample(times_resample)

            return times_resample, values_resample
        else:
            return times, values



from cadet import H5, Cadet
from addict import Dict
import numpy

def set_discretization(model, defaults, n_bound=None):
    columns = {'GENERAL_RATE_MODEL', 'LUMPED_RATE_MODEL_WITH_PORES', 'LUMPED_RATE_MODEL_WITHOUT_PORES'}
    
    
    for unit_name, unit in model.root.input.model.items():
        if 'unit_' in unit_name and unit.unit_type in columns:
            unit.discretization.ncol = defaults.ncol
            unit.discretization.npar = defaults.npar
            
            if n_bound is None:
                n_bound = unit.ncomp*[0]
            unit.discretization.nbound = n_bound
            
            unit.discretization.par_disc_type = 'EQUIDISTANT_PAR'
            unit.discretization.use_analytic_jacobian = 1
            unit.discretization.reconstruction = 'WENO'
            unit.discretization.gs_type = 1
            unit.discretization.max_krylov = 0
            unit.discretization.max_restarts = 10
            unit.discretization.schur_safety = 1.0e-8

            unit.discretization.weno.boundary_model = 0
            unit.discretization.weno.weno_eps = 1e-10
            unit.discretization.weno.weno_order = 3


def get_cadet_template(n_units=3, defaults=None):
    cadet_template = Cadet()
    
    cadet_template.root.input.model.nunits = n_units
    
    # Store solution
    cadet_template.root.input['return'].split_components_data = 1
    cadet_template.root.input['return'].unit_000.write_solution_inlet = 1
    cadet_template.root.input['return'].unit_000.write_solution_outlet = 1
    
    for unit in range(n_units):
        cadet_template.root.input['return']['unit_{0:03d}'.format(unit)] = cadet_template.root.input['return'].unit_000
        
    # Tolerances for the time integrator
    cadet_template.root.input.solver.time_integrator.abstol = defaults.abstol
    cadet_template.root.input.solver.time_integrator.algtol = defaults.algtol
    cadet_template.root.input.solver.time_integrator.reltol = defaults.reltol
    cadet_template.root.input.solver.time_integrator.init_step_size = 1e-6
    cadet_template.root.input.solver.time_integrator.max_steps = 1000000
    
    # Solver settings
    cadet_template.root.input.model.solver.gs_type = 1
    cadet_template.root.input.model.solver.max_krylov = 0
    cadet_template.root.input.model.solver.max_restarts = 10
    cadet_template.root.input.model.solver.schur_safety = 1e-8

    # Run the simulation on single thread
    cadet_template.root.input.solver.nthreads = 0
    
    return cadet_template


def create_dextran_model(defaults):

    dextran_model = get_cadet_template(n_units=3, defaults=defaults)

    # INLET
    dextran_model.root.input.model.unit_000.unit_type = 'INLET'
    dextran_model.root.input.model.unit_000.ncomp = 1
    dextran_model.root.input.model.unit_000.inlet_type = 'PIECEWISE_CUBIC_POLY'

    # Column
    dextran_model.root.input.model.unit_001.unit_type = 'LUMPED_RATE_MODEL_WITH_PORES'
    dextran_model.root.input.model.unit_001.ncomp = 1
    
    dextran_model.root.input.model.unit_001.col_length = 0.25
    dextran_model.root.input.model.unit_001.cross_section_area = 1.0386890710931253E-4
    dextran_model.root.input.model.unit_001.col_porosity = 0.37
    dextran_model.root.input.model.unit_001.par_porosity = 0.33
    dextran_model.root.input.model.unit_001.par_radius = 4.5e-5

    dextran_model.root.input.model.unit_001.col_dispersion = defaults.col_dispersion
    dextran_model.root.input.model.unit_001.film_diffusion = [0.0,]
    
    dextran_model.root.input.model.unit_001.adsorption_model = 'NONE'
    
    dextran_model.root.input.model.unit_001.init_c = [0.0,]
        
    set_discretization(dextran_model, defaults)
    
    ## Outlet
    dextran_model.root.input.model.unit_002.ncomp = 1
    dextran_model.root.input.model.unit_002.unit_type = 'OUTLET'
    
    # Sections and connections
    dextran_model.root.input.solver.sections.nsec = 2
    dextran_model.root.input.solver.sections.section_times = [0.0, 12.0, 600.0]
    dextran_model.root.input.solver.sections.section_continuity = [0,]
    
    ## Inlet Profile
    dextran_model.root.input.model.unit_000.sec_000.const_coeff = [0.002,]
    dextran_model.root.input.model.unit_000.sec_000.lin_coeff = [0.0,]
    dextran_model.root.input.model.unit_000.sec_000.quad_coeff = [0.0,]
    dextran_model.root.input.model.unit_000.sec_000.cube_coeff = [0.0,]


    dextran_model.root.input.model.unit_000.sec_001.const_coeff = [0.0,]
    dextran_model.root.input.model.unit_000.sec_001.lin_coeff = [0.0,]
    dextran_model.root.input.model.unit_000.sec_001.quad_coeff = [0.0,]
    dextran_model.root.input.model.unit_000.sec_001.cube_coeff = [0.0,]
    
    ## Switches
    dextran_model.root.input.model.connections.nswitches = 1
    dextran_model.root.input.model.connections.switch_000.section = 0
    dextran_model.root.input.model.connections.switch_000.connections = [
        0, 1, -1, -1, defaults.flow_rate,
        1, 2, -1, -1, defaults.flow_rate
    ]

    #set the times that the simulator writes out data for
    dextran_model.root.input.solver.user_solution_times = numpy.linspace(0, 600, 601)

    return dextran_model

def create_nonbinding_model(defaults):
    dextran_model = create_dextran_model(defaults)
    dextran_model.root.input.model.unit_001.film_diffusion = [defaults.film_diffusion,]
    dextran_model.root.input.solver.sections.section_times = [0.0, 12.0, 1000.0]
    dextran_model.root.input.solver.user_solution_times = numpy.linspace(0, 1000, 1001)

    return dextran_model

def create_linear_model(defaults):

    linear_model = get_cadet_template(n_units=3, defaults=defaults)

    # INLET
    linear_model.root.input.model.unit_000.unit_type = 'INLET'
    linear_model.root.input.model.unit_000.ncomp = 2
    linear_model.root.input.model.unit_000.inlet_type = 'PIECEWISE_CUBIC_POLY'

    # Column
    linear_model.root.input.model.unit_001.unit_type = 'GENERAL_RATE_MODEL'
    linear_model.root.input.model.unit_001.ncomp = 2
    
    linear_model.root.input.model.unit_001.col_length = 0.25
    linear_model.root.input.model.unit_001.cross_section_area = 1.0386890710931253E-4
    linear_model.root.input.model.unit_001.col_porosity = 0.37
    linear_model.root.input.model.unit_001.par_porosity = 0.33
    linear_model.root.input.model.unit_001.par_radius = 4.5e-5

    linear_model.root.input.model.unit_001.col_dispersion = defaults.col_dispersion
    linear_model.root.input.model.unit_001.film_diffusion = [defaults.film_diffusion,defaults.film_diffusion]
    linear_model.root.input.model.unit_001.par_diffusion = [defaults.par_diffusion, defaults.par_diffusion]
    
    linear_model.root.input.model.unit_001.adsorption_model = 'LINEAR'

    linear_model.root.input.model.unit_001.adsorption.is_kinetic = 1
    linear_model.root.input.model.unit_001.adsorption.lin_ka = [defaults.lin_ka1, defaults.lin_ka2]
    linear_model.root.input.model.unit_001.adsorption.lin_kd = [defaults.lin_kd1, defaults.lin_kd2]
    
    linear_model.root.input.model.unit_001.init_c = [0.0,0.0]
    linear_model.root.input.model.unit_001.init_q = [0.0,0.0]
        
    set_discretization(linear_model, defaults, n_bound=[1,1])
    
    ## Outlet
    linear_model.root.input.model.unit_002.ncomp = 2
    linear_model.root.input.model.unit_002.unit_type = 'OUTLET'
    
    # Sections and connections
    linear_model.root.input.solver.sections.nsec = 2
    linear_model.root.input.solver.sections.section_times = [0.0, 50.0, 1200.0]
    linear_model.root.input.solver.sections.section_continuity = [0,]
    
    ## Inlet Profile
    linear_model.root.input.model.unit_000.sec_000.const_coeff = [0.1,0.1]
    linear_model.root.input.model.unit_000.sec_000.lin_coeff = [0.0,0.0]
    linear_model.root.input.model.unit_000.sec_000.quad_coeff = [0.0,0.0]
    linear_model.root.input.model.unit_000.sec_000.cube_coeff = [0.0,0.0]

    linear_model.root.input.model.unit_000.sec_001.const_coeff = [0.0,0.0]
    linear_model.root.input.model.unit_000.sec_001.lin_coeff = [0.0,0.0]
    linear_model.root.input.model.unit_000.sec_001.quad_coeff = [0.0,0.0]
    linear_model.root.input.model.unit_000.sec_001.cube_coeff = [0.0,0.0]
    
    ## Switches
    linear_model.root.input.model.connections.nswitches = 1
    linear_model.root.input.model.connections.switch_000.section = 0
    linear_model.root.input.model.connections.switch_000.connections = [
        0, 1, -1, -1, defaults.flow_rate,
        1, 2, -1, -1, defaults.flow_rate
    ]

    #set the times that the simulator writes out data for
    linear_model.root.input.solver.user_solution_times = numpy.linspace(0, 1200, 1201)

    return linear_model

def create_cstr_model(defaults):

    cstr_model = get_cadet_template(n_units=5, defaults=defaults)

    # INLET
    cstr_model.root.input.model.unit_000.unit_type = 'INLET'
    cstr_model.root.input.model.unit_000.ncomp = 1
    cstr_model.root.input.model.unit_000.inlet_type = 'PIECEWISE_CUBIC_POLY'

    # CSTR
    cstr_model.root.input.model.unit_001.unit_type = 'CSTR'
    cstr_model.root.input.model.unit_001.ncomp = 1
    cstr_model.root.input.model.unit_001.init_c = [0.0]
    cstr_model.root.input.model.unit_001.adsorption_model = 'NONE'
    cstr_model.root.input.model.unit_001.init_volume = 5e-6

    cstr_model.root.input.model.unit_003.unit_type = 'CSTR'
    cstr_model.root.input.model.unit_003.ncomp = 1
    cstr_model.root.input.model.unit_003.init_c = [0.0]
    cstr_model.root.input.model.unit_003.adsorption_model = 'NONE'
    cstr_model.root.input.model.unit_003.init_volume = 5e-6

    cstr_model.root.input.model.unit_004.unit_type = 'CSTR'
    cstr_model.root.input.model.unit_004.ncomp = 1
    cstr_model.root.input.model.unit_004.init_c = [0.0]
    cstr_model.root.input.model.unit_004.adsorption_model = 'NONE'
    cstr_model.root.input.model.unit_004.init_volume = 1e-5

    
    ## Outlet
    cstr_model.root.input.model.unit_002.ncomp = 1
    cstr_model.root.input.model.unit_002.unit_type = 'OUTLET'
    
    # Sections and connections
    cstr_model.root.input.solver.sections.nsec = 2
    cstr_model.root.input.solver.sections.section_times = [0.0, 50.0, 1200.0]
    cstr_model.root.input.solver.sections.section_continuity = [0,]
    
    ## Inlet Profile
    cstr_model.root.input.model.unit_000.sec_000.const_coeff = [0.1]
    cstr_model.root.input.model.unit_000.sec_000.lin_coeff = [0.0,]
    cstr_model.root.input.model.unit_000.sec_000.quad_coeff = [0.0,]
    cstr_model.root.input.model.unit_000.sec_000.cube_coeff = [0.0,]

    cstr_model.root.input.model.unit_000.sec_001.const_coeff = [0.0]
    cstr_model.root.input.model.unit_000.sec_001.lin_coeff = [0.0,]
    cstr_model.root.input.model.unit_000.sec_001.quad_coeff = [0.0,]
    cstr_model.root.input.model.unit_000.sec_001.cube_coeff = [0.0,]
    
    ## Switches
    cstr_model.root.input.model.connections.nswitches = 1
    cstr_model.root.input.model.connections.switch_000.section = 0
    cstr_model.root.input.model.connections.switch_000.connections = [
        0, 1, -1, -1, defaults.flow_rate,
        1, 3, -1, -1, defaults.flow_rate,
        3, 2, -1, -1, defaults.flow_rate,
        0, 4, -1, -1, defaults.flow_rate,
        4, 2, -1, -1, defaults.flow_rate
    ]

    #set the times that the simulator writes out data for
    cstr_model.root.input.solver.user_solution_times = numpy.linspace(0, 1200, 1201*10)

    return cstr_model



import numpy

import CADETMatch.util as util
from CADETMatch.abstract.transform import AbstractTransform


class VolumeLengthTransform(AbstractTransform):
    @property
    def name(self):
        return "volume_length"

    @property
    def count(self):
        return 2

    @property
    def count_extended(self):
        return 3

    def transform(self):
        def trans_volume(i):
            return i

        def trans_length(i):
            return i

        return [trans_volume, trans_length]

    grad_transform = transform

    def untransform_inputorder(self, seq):
        volume = seq[0]
        length = seq[1]
        return [volume, length]

    def untransform(self, seq):
        volume, length = self.untransform_inputorder(seq)
        area = volume / length
        values = [area, length]
        headerValues = [area, length, volume]
        return values, headerValues

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        values = self.untransform_matrix_inputorder(matrix)
        values[:, 0] = values[:, 0] / values[:, 1]
        return values

    def untransform_matrix_inputorder(self, matrix):
        values = numpy.zeros(matrix.shape)
        values[:, 0] = matrix[:, 0]
        values[:, 1] = matrix[:, 1]
        return values

    def setSimulation(self, sim, seq, experiment):
        values, headerValues = self.untransform(seq)

        if (
            self.parameter.get("experiments", None) is None
            or experiment["name"] in self.parameter["experiments"]
        ):
            area_location = self.parameter["area_location"]
            length_location = self.parameter["length_location"]

            sim[area_location.lower()] = values[0]
            sim[length_location.lower()] = values[1]

        return values, headerValues

    def getBounds(self):
        minVolume = self.parameter["minVolume"]
        maxVolume = self.parameter["maxVolume"]
        minLength = self.parameter["minLength"]
        maxLength = self.parameter["maxLength"]

        minValues = [minVolume, minLength]
        maxValues = [maxVolume, maxLength]

        return minValues, maxValues

    def getGradBounds(self):
        return self.getBounds()

    def getHeaders(self):
        headers = []
        headers.append("Area")
        headers.append("Length")
        headers.append("Volume")
        return headers

    def getHeadersActual(self):
        headers = []
        headers.append("Volume")
        headers.append("Length")
        return headers

    def setBounds(self, parameter, lb, ub):
        parameter["minVolume"] = lb[0]
        parameter["maxVolume"] = ub[0]
        parameter["minLength"] = lb[1]
        parameter["maxLength"] = ub[1]


class NormVolumeLengthTransform(VolumeLengthTransform):
    @property
    def name(self):
        return "norm_volume_length"

    def transform(self):
        minVolume = self.parameter["minVolume"]
        maxVolume = self.parameter["maxVolume"]
        minLength = self.parameter["minLength"]
        maxLength = self.parameter["maxLength"]

        def trans_volume(i):
            return (i - minVolume) / (maxVolume - minVolume)

        def trans_length(i):
            return (i - minLength) / (maxLength - minLength)

        return [trans_volume, trans_length]

    grad_transform = transform

    def untransform_inputorder(self, seq):
        minVolume = self.parameter["minVolume"]
        maxVolume = self.parameter["maxVolume"]
        minLength = self.parameter["minLength"]
        maxLength = self.parameter["maxLength"]

        minValues = numpy.array([minVolume, minLength])
        maxValues = numpy.array([maxVolume, maxLength])

        values = numpy.array(seq)

        values = (maxValues - minValues) * values + minValues
        return values

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        values = self.untransform_matrix_inputorder(matrix)
        values[:, 0] = values[:, 0] / values[:, 1]
        return values

    def untransform_matrix_inputorder(self, matrix):
        minVolume = self.parameter["minVolume"]
        maxVolume = self.parameter["maxVolume"]
        minLength = self.parameter["minLength"]
        maxLength = self.parameter["maxLength"]

        minValues = numpy.array([minVolume, minLength])
        maxValues = numpy.array([maxVolume, maxLength])

        values = (maxValues - minValues) * matrix + minValues

        return values

    def getBounds(self):
        return [0.0, 0.0], [1.0, 1.0]

    def getGradBounds(self):
        return [self.parameter["min"],], [
            self.parameter["max"],
        ]


plugins = {
    "norm_volume_length": NormVolumeLengthTransform,
    "volume_length": VolumeLengthTransform,
}
